import pymc as pm
from pymc import six
import numpy as np


class TruncatedMetropolis(pm.Metropolis):

    def __init__(self, stochastic, low_bound, up_bound, *args, **kwargs):
        self.low_bound = low_bound
        self.up_bound = up_bound
        pm.Metropolis.__init__(self, stochastic, *args, **kwargs)

    # Propose method written by hacking Metropolis.propose()
    def propose(self):
        tau = 1. / (self.adaptive_scale_factor * self.proposal_sd) ** 2
        self.stochastic.value = pm.rtruncnorm(
            self.stochastic.value,
            tau,
            self.low_bound,
            self.up_bound)

    # Hastings factor method accounts for asymmetric proposal distribution
    def hastings_factor(self):
        tau = 1. / (self.adaptive_scale_factor * self.proposal_sd) ** 2
        cur_val = self.stochastic.value
        last_val = self.stochastic.last_value

        lp_for = pm.truncnorm_like(
            cur_val,
            last_val,
            tau,
            self.low_bound,
            self.up_bound)
        lp_bak = pm.truncnorm_like(
            last_val,
            cur_val,
            tau,
            self.low_bound,
            self.up_bound)

        if self.verbose > 1:
            six.print_(self._id + ': Hastings factor %f' % (lp_bak - lp_for))
        return lp_bak - lp_for


# Maximum data value is around 1 (plot the histogram).
data = np.array(
    [-1.6464815, -0.86463278, 0.80656378, 0.67664181, -0.34312965,
     -0.29654303, 0.76170081, 0.30418603, -0.45473639, -0.24894277,
     -0.07173209, -1.64602289, 0.0804062, -0.82159472, 0.98224623,
     -1.92538425, -1.95388748, -0.41145515, 0.23972844, -0.78645389,
     -0.21687104, -0.2939634, 0.51229013, 0.04626286, 0.18329919,
     -1.12775839, -1.64187249, 0.33440094, -0.95224695, 0.15650266,
     -0.54056102, 0.12240128, -0.95397459, 0.44806432, -1.02955556,
        0.31740861, -0.8762523, 0.47377688, 0.76516415, 0.27890419,
     -0.07819642, -0.13399348, 0.82877293, 0.22308624, 0.7485783,
     -0.14700254, -1.03145657, 0.85641097, 0.43396285, 0.47901653,
        0.80137086, 0.33566812, 0.71443253, -1.57590815, -0.24090179,
     -2.0128344, 0.34503324, 0.12944091, -1.5327008, 0.06363034,
        0.21042021, -0.81425636, 0.20209279, -1.48130423, -1.04983523,
        0.16001774, -0.75239072, 0.33427956, -0.10224921, 0.26463561,
     -1.09374674, -0.72749811, -0.54892116, -1.89631844, -0.94393545,
     -0.2521341, 0.26840341, 0.23563219, 0.35333094])

# Model: the data are truncated-normally distributed with unknown upper bound.
mu = pm.Normal('mu', 0, .01, value=0)
tau = pm.Exponential('tau', .01, value=1)
cutoff = pm.Exponential('cutoff', 1, value=1.3)
D = pm.TruncatedNormal(
    'D',
    mu,
    tau,
    -np.inf,
    cutoff,
    value=data,
    observed=True)

M = pm.MCMC([mu, tau, cutoff, D])

# Use a TruncatedMetropolis step method that will never propose jumps
# below D's maximum value.
M.use_step_method(TruncatedMetropolis, cutoff, D.value.max(), np.inf)
# Get a handle to the step method handling cutoff to investigate its behavior.
S = M.step_method_dict[cutoff][0]

M.isample(10000, 0, 10)



# Import the mean and covariance
from .mean import M
from .cov import C
from pymc.gp import *

# Generate realizations
f_list = [Realization(M, C) for i in range(3)]

# - Plot - ####
if __name__ == '__main__':
    from pylab import *

    x = arange(-1., 1., .01)
    clf()

    plot_envelope(M, C, x)

    for f in f_list:
        plot(x, f(x))

    xlabel('x')
    ylabel('f(x)')
    title('Three realizations of the GP')
    axis('tight')

    # show()



"""
A model for the disasters data with a changepoint, with GOF diagnostics added

changepoint ~ U(0,110)
early_mean ~ Exp(1.)
late_mean ~ Exp(1.)
disasters[t] ~ Po(early_mean if t <= switchpoint, late_mean otherwise)

"""

import pymc as pm
from numpy import array, concatenate, ones
from numpy.random import randint

__all__ = [
    'disasters_array',
    'switchpoint',
    'early_mean',
    'late_mean',
    'disasters']

disasters_array = array([4, 5, 4, 0, 1, 4, 3, 4, 0, 6, 3, 3, 4, 0, 2, 6,
                         3, 3, 5, 4, 5, 3, 1, 4, 4, 1, 5, 5, 3, 4, 2, 5,
                         2, 2, 3, 4, 2, 1, 3, 2, 2, 1, 1, 1, 1, 3, 0, 0,
                         1, 0, 1, 1, 0, 0, 3, 1, 0, 3, 2, 2, 0, 1, 1, 1,
                         0, 1, 0, 1, 0, 0, 0, 2, 1, 0, 0, 0, 1, 1, 0, 2,
                         3, 3, 1, 1, 2, 1, 1, 1, 1, 2, 4, 2, 0, 0, 1, 4,
                         0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1])

n = len(disasters_array)

# Define data and stochastics

switchpoint = pm.DiscreteUniform('switchpoint', lower=0, upper=110)
early_mean = pm.Exponential('early_mean', beta=1.)
late_mean = pm.Exponential('late_mean', beta=1.)


@pm.stochastic(observed=True, dtype=int)
def disasters(value=disasters_array,
              early_mean=early_mean,
              late_mean=late_mean,
              switchpoint=switchpoint):
    """Annual occurences of coal mining disasters."""
    return pm.poisson_like(
        value[:switchpoint], early_mean) + pm.poisson_like(value[switchpoint:], late_mean)


@pm.deterministic
def disasters_sim(early_mean=early_mean,
                  late_mean=late_mean,
                  switchpoint=switchpoint):
    """Coal mining disasters sampled from the posterior predictive distribution"""
    return concatenate((pm.rpoisson(early_mean, size=switchpoint), pm.rpoisson(
        late_mean, size=n - switchpoint)))


@pm.deterministic
def expected_values(early_mean=early_mean,
                    late_mean=late_mean,
                    switchpoint=switchpoint):
    """Discrepancy measure for GOF using the Freeman-Tukey statistic"""

    # Sample size
    n = len(disasters_array)
    # Expected values
    return concatenate(
        (ones(switchpoint) * early_mean, ones(n - switchpoint) * late_mean))


if __name__ == '__main__':
    vars = [
        switchpoint,
        early_mean,
        late_mean,
        disasters,
        disasters_sim,
        expected_values]
    # Instiatiate model
    M = pm.MCMC(vars)
    # Sample
    M.sample(10000, burn=5000, verbose=2)
    # Calculate discrepancy function
    D = pm.diagnostics.discrepancy(
        disasters_array,
        disasters_sim,
        expected_values)
    # Plot GOF graphics
    pm.Matplot.discrepancy_plot(D, 'D')
    pm.Matplot.gof_plot(disasters_sim, disasters_array, 'disasters')



# Convergence diagnostics and model validation
import numpy as np
from . import Matplot, flib
from .utils import autocorr, autocov
from copy import copy
import pdb

try:
    from statsmodels.regression.linear_model import yule_walker
    has_sm = True
except ImportError:
    has_sm = False

from . import six
from .six import print_
xrange = six.moves.xrange

__all__ = [
    'geweke',
    'gelman_rubin',
    'raftery_lewis',
    'validate',
    'discrepancy',
    'iat',
    'ppp_value',
    'effective_n']


def open01(x, limit=1.e-6):
    """Constrain numbers to (0,1) interval"""
    try:
        return np.array([min(max(y, limit), 1. - limit) for y in x])
    except TypeError:
        return min(max(x, limit), 1. - limit)


class diagnostic(object):

    """
    This decorator allows for PyMC arguments of various types to be passed to
    the diagnostic functions. It identifies the type of object and locates its
    trace(s), then passes the data to the wrapped diagnostic function.

    """

    def __init__(self, all_chains=False):
        """ Initialize wrapper """

        self.all_chains = all_chains

    def __call__(self, f):

        def wrapped_f(pymc_obj, *args, **kwargs):

            # Figure out what type of object it is
            try:
                values = {}
                # First try Model type
                for variable in pymc_obj._variables_to_tally:
                    if self.all_chains:
                        k = pymc_obj.db.chains
                        data = [variable.trace(chain=i) for i in range(k)]
                    else:
                        data = variable.trace()
                    name = variable.__name__
                    if kwargs.get('verbose'):
                        print_("\nDiagnostic for %s ..." % name)
                    values[name] = f(data, *args, **kwargs)
                return values
            except AttributeError:
                pass

            try:
                # Then try Node type
                if self.all_chains:
                    k = pymc_obj.trace.db.chains
                    data = [pymc_obj.trace(chain=i) for i in range(k)]
                else:
                    data = pymc_obj.trace()
                name = pymc_obj.__name__
                return f(data, *args, **kwargs)
            except (AttributeError, ValueError):
                pass

            # If others fail, assume that raw data is passed
            return f(pymc_obj, *args, **kwargs)

        wrapped_f.__doc__ = f.__doc__
        wrapped_f.__name__ = f.__name__

        return wrapped_f


def validate(sampler, replicates=20, iterations=10000, burn=5000,
             thin=1, deterministic=False, db='ram', plot=True, verbose=0):
    """
    Model validation method, following Cook et al. (Journal of Computational and
    Graphical Statistics, 2006, DOI: 10.1198/106186006X136976).

    Generates posterior samples based on 'true' parameter values and data simulated
    from the priors. The quantiles of the parameter values are calculated, based on
    the samples. If the model is valid, the quantiles should be uniformly distributed
    over [0,1].

    Since this relies on the generation of simulated data, all data stochastics
    must have a valid random() method for validation to proceed.

    Parameters
    ----------
    sampler : Sampler
      An MCMC sampler object.
    replicates (optional) : int
      The number of validation replicates (i.e. number of quantiles to be simulated).
      Defaults to 100.
    iterations (optional) : int
      The number of MCMC iterations to be run per replicate. Defaults to 2000.
    burn (optional) : int
      The number of burn-in iterations to be run per replicate. Defaults to 1000.
    thin (optional) : int
      The thinning factor to be applied to posterior sample. Defaults to 1 (no thinning)
    deterministic (optional) : bool
      Flag for inclusion of deterministic nodes in validation procedure. Defaults
      to False.
    db (optional) : string
      The database backend to use for the validation runs. Defaults to 'ram'.
    plot (optional) : bool
      Flag for validation plots. Defaults to True.

    Returns
    -------
    stats : dict
      Return a dictionary containing tuples with the chi-square statistic and
      associated p-value for each data stochastic.

    Notes
    -----
    This function requires SciPy.
    """
    import scipy as sp
    # Set verbosity for models to zero
    sampler.verbose = 0

    # Specify parameters to be evaluated
    parameters = sampler.stochastics
    if deterministic:
        # Add deterministics to the mix, if requested
        parameters = parameters | sampler.deterministics

    # Assign database backend
    original_backend = sampler.db.__name__
    sampler._assign_database_backend(db)

    # Empty lists for quantiles
    quantiles = {}

    if verbose:
        print_("\nExecuting Cook et al. (2006) validation procedure ...\n")

    # Loop over replicates
    for i in range(replicates):

        # Sample from priors
        for p in sampler.stochastics:
            if not p.extended_parents:
                p.random()

        # Sample "true" data values
        for o in sampler.observed_stochastics:
            # Generate simuated data for data stochastic
            o.set_value(o.random(), force=True)
            if verbose:
                print_("Data for %s is %s" % (o.__name__, o.value))

        param_values = {}
        # Record data-generating parameter values
        for s in parameters:
            param_values[s] = s.value

        try:
            # Fit models given parameter values
            sampler.sample(iterations, burn=burn, thin=thin)

            for s in param_values:

                if not i:
                    # Initialize dict
                    quantiles[s.__name__] = []
                trace = s.trace()
                q = sum(trace < param_values[s], 0) / float(len(trace))
                quantiles[s.__name__].append(open01(q))

            # Replace data values
            for o in sampler.observed_stochastics:
                o.revert()

        finally:
            # Replace data values
            for o in sampler.observed_stochastics:
                o.revert()

            # Replace backend
            sampler._assign_database_backend(original_backend)

        if not i % 10 and i and verbose:
            print_("\tCompleted validation replicate", i)

    # Replace backend
    sampler._assign_database_backend(original_backend)

    stats = {}
    # Calculate chi-square statistics
    for param in quantiles:
        q = quantiles[param]
        # Calculate chi-square statistics
        X2 = sum(sp.special.ndtri(q) ** 2)
        # Calculate p-value
        p = sp.special.chdtrc(replicates, X2)

        stats[param] = (X2, p)

    if plot:
        # Convert p-values to z-scores
        p = copy(stats)
        for i in p:
            p[i] = p[i][1]
        Matplot.zplot(p, verbose=verbose)

    return stats

def spec(x, order=2):
    
    beta, sigma = yule_walker(x, order)
    return sigma**2 / (1. - np.sum(beta))**2

@diagnostic()
def geweke(x, first=.1, last=.5, intervals=20, maxlag=20):
    """Return z-scores for convergence diagnostics.

    Compare the mean of the first % of series with the mean of the last % of
    series. x is divided into a number of segments for which this difference is
    computed. If the series is converged, this score should oscillate between
    -1 and 1.

    Parameters
    ----------
    x : array-like
      The trace of some stochastic parameter.
    first : float
      The fraction of series at the beginning of the trace.
    last : float
      The fraction of series at the end to be compared with the section
      at the beginning.
    intervals : int
      The number of segments.
    maxlag : int
      Maximum autocorrelation lag for estimation of spectral variance

    Returns
    -------
    scores : list [[]]
      Return a list of [i, score], where i is the starting index for each
      interval and score the Geweke score on the interval.

    Notes
    -----

    The Geweke score on some series x is computed by:

      .. math:: \frac{E[x_s] - E[x_e]}{\sqrt{V[x_s] + V[x_e]}}

    where :math:`E` stands for the mean, :math:`V` the variance,
    :math:`x_s` a section at the start of the series and
    :math:`x_e` a section at the end of the series.

    References
    ----------
    Geweke (1992)
    """
    
    if not has_sm:
        print("statsmodels not available. Geweke diagnostic cannot be calculated.")
        return

    if np.ndim(x) > 1:
        return [geweke(y, first, last, intervals) for y in np.transpose(x)]

    # Filter out invalid intervals
    if first + last >= 1:
        raise ValueError(
            "Invalid intervals for Geweke convergence analysis",
            (first, last))

    # Initialize list of z-scores
    zscores = [None] * intervals

    # Starting points for calculations
    starts = np.linspace(0, int(len(x)*(1.-last)), intervals).astype(int)

    # Loop over start indices
    for i,s in enumerate(starts):

        # Size of remaining array
        x_trunc = x[s:]
        n = len(x_trunc)

        # Calculate slices
        first_slice = x_trunc[:int(first * n)]
        last_slice = x_trunc[int(last * n):]

        z = (first_slice.mean() - last_slice.mean())
        z /= np.sqrt(spec(first_slice)/len(first_slice) +
                     spec(last_slice)/len(last_slice))
        zscores[i] = len(x) - n, z

    return zscores

# From StatLib -- gibbsit.f


@diagnostic()
def raftery_lewis(x, q, r, s=.95, epsilon=.001, verbose=1):
    """
    Return the number of iterations needed to achieve a given
    precision.

    :Parameters:
        x : sequence
            Sampled series.
        q : float
            Quantile.
        r : float
            Accuracy requested for quantile.
        s (optional) : float
            Probability of attaining the requested accuracy (defaults to 0.95).
        epsilon (optional) : float
             Half width of the tolerance interval required for the q-quantile (defaults to 0.001).
        verbose (optional) : int
            Verbosity level for output (defaults to 1).

    :Return:
        nmin : int
            Minimum number of independent iterates required to achieve
            the specified accuracy for the q-quantile.
        kthin : int
            Skip parameter sufficient to produce a first-order Markov
            chain.
        nburn : int
            Number of iterations to be discarded at the beginning of the
            simulation, i.e. the number of burn-in iterations.
        nprec : int
            Number of iterations not including the burn-in iterations which
            need to be obtained in order to attain the precision specified
            by the values of the q, r and s input parameters.
        kmind : int
            Minimum skip parameter sufficient to produce an independence
            chain.

    :Example:
        >>> raftery_lewis(x, q=.025, r=.005)

    :Reference:
        Raftery, A.E. and Lewis, S.M. (1995).  The number of iterations,
        convergence diagnostics and generic Metropolis algorithms.  In
        Practical Markov Chain Monte Carlo (W.R. Gilks, D.J. Spiegelhalter
        and S. Richardson, eds.). London, U.K.: Chapman and Hall.

        See the fortran source file `gibbsit.f` for more details and references.
    """
    if np.ndim(x) > 1:
        return [raftery_lewis(y, q, r, s, epsilon, verbose)
                for y in np.transpose(x)]

    output = nmin, kthin, nburn, nprec, kmind = flib.gibbmain(
        x, q, r, s, epsilon)

    if verbose:

        print_("\n========================")
        print_("Raftery-Lewis Diagnostic")
        print_("========================")
        print_()
        print_(
            "%s iterations required (assuming independence) to achieve %s accuracy with %i percent probability." %
            (nmin, r, 100 * s))
        print_()
        print_(
            "Thinning factor of %i required to produce a first-order Markov chain." %
            kthin)
        print_()
        print_(
            "%i iterations to be discarded at the beginning of the simulation (burn-in)." %
            nburn)
        print_()
        print_("%s subsequent iterations required." % nprec)
        print_()
        print_(
            "Thinning factor of %i required to produce an independence chain." %
            kmind)

    return output


def batch_means(x, f=lambda y: y, theta=.5, q=.95, burn=0):
    """
    TODO: Use Bayesian CI.

    Returns the half-width of the frequentist confidence interval
    (q'th quantile) of the Monte Carlo estimate of E[f(x)].

    :Parameters:
        x : sequence
            Sampled series. Must be a one-dimensional array.
        f : function
            The MCSE of E[f(x)] will be computed.
        theta : float between 0 and 1
            The batch length will be set to len(x) ** theta.
        q : float between 0 and 1
            The desired quantile.

    :Example:
        >>>batch_means(x, f=lambda x: x**2, theta=.5, q=.95)

    :Reference:
        Flegal, James M. and Haran, Murali and Jones, Galin L. (2007).
        Markov chain Monte Carlo: Can we trust the third significant figure?
        <Publication>

    :Note:
        Requires SciPy
    """

    try:
        import scipy
        from scipy import stats
    except ImportError:
        raise ImportError('SciPy must be installed to use batch_means.')

    x = x[burn:]

    n = len(x)

    b = np.int(n ** theta)
    a = n / b

    t_quant = stats.t.isf(1 - q, a - 1)

    Y = np.array([np.mean(f(x[i * b:(i + 1) * b])) for i in xrange(a)])
    sig = b / (a - 1.) * sum((Y - np.mean(f(x))) ** 2)

    return t_quant * sig / np.sqrt(n)


def discrepancy(observed, simulated, expected):
    """Calculates Freeman-Tukey statistics (Freeman and Tukey 1950) as
    a measure of discrepancy between observed and r replicates of simulated data. This
    is a convenient method for assessing goodness-of-fit (see Brooks et al. 2000).

    D(x|\theta) = \sum_j (\sqrt{x_j} - \sqrt{e_j})^2

    :Parameters:
      observed : Iterable of observed values (size=(n,))
      simulated : Iterable of simulated values (size=(r,n))
      expected : Iterable of expected values (size=(r,) or (r,n))

    :Returns:
      D_obs : Discrepancy of observed values
      D_sim : Discrepancy of simulated values

    """
    try:
        simulated = simulated.astype(float)
    except AttributeError:
        simulated = simulated.trace().astype(float)
    try:
        expected = expected.astype(float)
    except AttributeError:
        expected = expected.trace().astype(float)
    # Ensure expected values are rxn
    expected = np.resize(expected, simulated.shape)

    D_obs = np.sum([(np.sqrt(observed) - np.sqrt(
        e)) ** 2 for e in expected], 1)
    D_sim = np.sum(
        [(np.sqrt(s) - np.sqrt(e)) ** 2 for s,
         e in zip(simulated,
                  expected)],
        1)

    # Print p-value
    count = sum(s > o for o, s in zip(D_obs, D_sim))
    print_('Bayesian p-value: p=%.3f' % (1. * count / len(D_obs)))

    return D_obs, D_sim
    

@diagnostic(all_chains=True)
def effective_n(x):
    """ Returns estimate of the effective sample size of a set of traces.

    Parameters
    ----------
    x : array-like
      An array containing the 2 or more traces of a stochastic parameter. That is, an array of dimension m x n x k, where m is the number of traces, n the number of samples, and k the dimension of the stochastic.
    
    Returns
    -------
    n_eff : float
      Return the effective sample size, :math:`\hat{n}_{eff}`

    Notes
    -----

    The diagnostic is computed by:

      .. math:: \hat{n}_{eff} = \frac{mn}}{1 + 2 \sum_{t=1}^T \hat{\rho}_t}

    where :math:`\hat{\rho}_t` is the estimated autocorrelation at lag t, and T
    is the first odd positive integer for which the sum :math:`\hat{\rho}_{T+1} + \hat{\rho}_{T+1}` 
    is negative.

    References
    ----------
    Gelman et al. (2014)"""
    
    if np.shape(x) < (2,):
        raise ValueError(
            'Calculation of effective sample size requires multiple chains of the same length.')

    try:
        m, n = np.shape(x)
    except ValueError:
        return [effective_n(np.transpose(y)) for y in np.transpose(x)]
        
    s2 = gelman_rubin(x, return_var=True)
    
    negative_autocorr = False
    t = 1
    
    variogram = lambda t: (sum(sum((x[j][i] - x[j][i-t])**2 for i in range(t,n)) for j in range(m)) 
                                / (m*(n - t)))
    rho = np.ones(n)
    # Iterate until the sum of consecutive estimates of autocorrelation is negative
    while not negative_autocorr and (t < n):
        
        rho[t] = 1. - variogram(t)/(2.*s2)
        
        if not t % 2:
            negative_autocorr = sum(rho[t-1:t+1]) < 0
        
        t += 1
        
    return int(m*n / (1 + 2*rho[1:t].sum()))
    

@diagnostic(all_chains=True)
def gelman_rubin(x, return_var=False):
    """ Returns estimate of R for a set of traces.

    The Gelman-Rubin diagnostic tests for lack of convergence by comparing
    the variance between multiple chains to the variance within each chain.
    If convergence has been achieved, the between-chain and within-chain
    variances should be identical. To be most effective in detecting evidence
    for nonconvergence, each chain should have been initialized to starting
    values that are dispersed relative to the target distribution.

    Parameters
    ----------
    x : array-like
      An array containing the 2 or more traces of a stochastic parameter. That is, an array of dimension m x n x k, where m is the number of traces, n the number of samples, and k the dimension of the stochastic.
      
    return_var : bool
      Flag for returning the marginal posterior variance instead of R-hat (defaults of False).

    Returns
    -------
    Rhat : float
      Return the potential scale reduction factor, :math:`\hat{R}`

    Notes
    -----

    The diagnostic is computed by:

      .. math:: \hat{R} = \sqrt{\frac{\hat{V}}{W}}

    where :math:`W` is the within-chain variance and :math:`\hat{V}` is
    the posterior variance estimate for the pooled traces. This is the
    potential scale reduction factor, which converges to unity when each
    of the traces is a sample from the target posterior. Values greater
    than one indicate that one or more chains have not yet converged.

    References
    ----------
    Brooks and Gelman (1998)
    Gelman and Rubin (1992)"""

    if np.shape(x) < (2,):
        raise ValueError(
            'Gelman-Rubin diagnostic requires multiple chains of the same length.')

    try:
        m, n = np.shape(x)
    except ValueError:
        return [gelman_rubin(np.transpose(y)) for y in np.transpose(x)]

    # Calculate between-chain variance
    B_over_n = np.sum((np.mean(x, 1) - np.mean(x)) ** 2) / (m - 1)

    # Calculate within-chain variances
    W = np.sum(
        [(x[i] - xbar) ** 2 for i,
         xbar in enumerate(np.mean(x,
                                   1))]) / (m * (n - 1))

    # (over) estimate of variance
    s2 = W * (n - 1) / n + B_over_n
    
    if return_var:
        return s2

    # Pooled posterior variance estimate
    V = s2 + B_over_n / m

    # Calculate PSRF
    R = V / W

    return np.sqrt(R)


def _find_max_lag(x, rho_limit=0.05, maxmaxlag=20000, verbose=0):
    """Automatically find an appropriate maximum lag to calculate IAT"""

    # Fetch autocovariance matrix
    acv = autocov(x)
    # Calculate rho
    rho = acv[0, 1] / acv[0, 0]

    lam = -1. / np.log(abs(rho))

    # Initial guess at 1.5 times lambda (i.e. 3 times mean life)
    maxlag = int(np.floor(3. * lam)) + 1

    # Jump forward 1% of lambda to look for rholimit threshold
    jump = int(np.ceil(0.01 * lam)) + 1

    T = len(x)

    while ((abs(rho) > rho_limit) & (maxlag < min(T / 2, maxmaxlag))):

        acv = autocov(x, maxlag)
        rho = acv[0, 1] / acv[0, 0]
        maxlag += jump

    # Add 30% for good measure
    maxlag = int(np.floor(1.3 * maxlag))

    if maxlag >= min(T / 2, maxmaxlag):
        maxlag = min(min(T / 2, maxlag), maxmaxlag)
        "maxlag fixed to %d" % maxlag
        return maxlag

    if maxlag <= 1:
        print_("maxlag = %d, fixing value to 10" % maxlag)
        return 10

    if verbose:
        print_("maxlag = %d" % maxlag)
    return maxlag


def _cut_time(gammas):
    """Support function for iat().
    Find cutting time, when gammas become negative."""

    for i in range(len(gammas) - 1):

        if not ((gammas[i + 1] > 0.0) & (gammas[i + 1] < gammas[i])):
            return i

    return i


@diagnostic()
def iat(x, maxlag=None):
    """Calculate the integrated autocorrelation time (IAT), given the trace from a Stochastic."""

    if not maxlag:
        # Calculate maximum lag to which autocorrelation is calculated
        maxlag = _find_max_lag(x)

    acr = [autocorr(x, lag) for lag in range(1, maxlag + 1)]

    # Calculate gamma values
    gammas = [(acr[2 * i] + acr[2 * i + 1]) for i in range(maxlag // 2)]

    cut = _cut_time(gammas)

    if cut + 1 == len(gammas):
        print_("Not enough lag to calculate IAT")

    return np.sum(2 * gammas[:cut + 1]) - 1.0


@diagnostic()
def ppp_value(simdata, trueval, round=3):
    """
    Calculates posterior predictive p-values on data simulated from the posterior
     predictive distribution, returning the quantile of the observed data relative to
     simulated.

     The posterior predictive p-value is computed by:

       .. math:: Pr(T(y^{\text{sim}} > T(y) | y)

     where T is a test statistic of interest and :math:`y^{\text{sim}}` is the simulated
     data.

    :Arguments:
        simdata: array or PyMC object
            Trace of simulated data or the PyMC stochastic object containing trace.

        trueval: numeric
            True (observed) value of the data

        round: int
            Rounding of returned quantile (defaults to 3)

    """

    if ndim(trueval) == 1 and ndim(simdata == 2):
        # Iterate over more than one set of data
        return [post_pred_checks(simdata[:, i], trueval[i])
                for i in range(len(trueval))]

    return (simdata > trueval).mean()



"""
Pickle backend module

Store the trace in a pickle file.

Notes
-----
Pickle file are not always compatible across different python
versions. Users should use this backend only for shortlived projects.

"""

from . import ram, no_trace, base
import os
import datetime
import numpy
import string

try:
    import cPickle as std_pickle
except ImportError:
    import pickle as std_pickle   # In Python 3, cPickle is folded into pickle

from pymc import six

__all__ = ['Trace', 'Database', 'load']


class Trace(ram.Trace):
    pass


class Database(base.Database):

    """Pickle database backend.

    Save the trace to a pickle file.
    """

    def __init__(self, dbname=None, dbmode='a'):
        """Assign a name to the file the database will be saved in.

        :Parameters:
        dbname : string
          Name of the pickle file.
        dbmode : {'a', 'w'}
          File mode.  Use `a` to append values, and `w` to overwrite
          an existing file.
        """
        self.__name__ = 'pickle'
        self.filename = dbname
        self.__Trace__ = Trace
        self.trace_names = []
        # A list of sequences of names of the objects to tally.
        self._traces = {}  # A dictionary of the Trace objects.
        self.chains = 0

        if os.path.exists(dbname):
            if dbmode == 'w':
                os.remove(dbname)

    def _finalize(self):
        """Dump traces using cPickle."""
        container = {}
        try:
            for name in self._traces:
                container[name] = self._traces[name]._trace
            container['_state_'] = self._state_

            file = open(self.filename, 'w+b')
            std_pickle.dump(container, file)
            file.close()
        except AttributeError:
            pass


def load(filename):
    """Load a pickled database.

    Return a Database instance.
    """
    file = open(filename, 'rb')
    container = std_pickle.load(file)
    file.close()
    db = Database(file.name)
    chains = 0
    funs = set()
    for k, v in six.iteritems(container):
        if k == '_state_':
            db._state_ = v
        else:
            db._traces[k] = Trace(name=k, value=v, db=db)
            setattr(db, k, db._traces[k])
            chains = max(chains, len(v))
            funs.add(k)

    db.chains = chains
    db.trace_names = chains * [list(funs)]

    return db



"""
A model for the disasters data with an linearly varying mean:

rate_of_mean ~ N(0,1)
amp_of_mean ~ Exp(3)
disasters[t] ~ Po(intercept_of_mean + slope_of_mean * t)
"""

from pymc import *
from numpy import *


disasters_array = array([4, 5, 4, 0, 1, 4, 3, 4, 0, 6, 3, 3, 4, 0, 2, 6,
                         3, 3, 5, 4, 5, 3, 1, 4, 4, 1, 5, 5, 3, 4, 2, 5,
                         2, 2, 3, 4, 2, 1, 3, 2, 2, 1, 1, 1, 1, 3, 0, 0,
                         1, 0, 1, 1, 0, 0, 3, 1, 0, 3, 2, 2, 0, 1, 1, 1,
                         0, 1, 0, 1, 0, 0, 0, 2, 1, 0, 0, 0, 1, 1, 0, 2,
                         3, 3, 1, 1, 2, 1, 1, 1, 1, 2, 4, 2, 0, 0, 1, 4,
                         0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1])

# Define data and stochastics


@stochastic
def params_of_mean(value=array([-.005, 1.]), tau=.1, rate=4.):
    """
    Intercept and slope of rate stochastic of poisson distribution
    Rate stochastic must be positive for t in [0,T]

    p(intercept, slope|tau,rate) =
    N(slope|0,tau) Exp(intercept|rate) 1(intercept>0) 1(intercept + slope * T>0)
    """

    def logp(value, tau, rate):
        if value[1] > 0 and value[1] + value[0] * 110 > 0:
            return normal_like(value[0], 0., tau) + \
                exponential_like(value[1], rate)
        else:
            return -Inf

    def random(tau, rate):
        val = zeros(2)
        val[0] = rnormal(0., tau)
        val[1] = rexponential(rate)
        while val[1] < 0 or val[1] + val[0] * 110 <= 0:
            val[0] = rnormal(0., tau)
            val[1] = rexponential(rate)
        return val


@observed
def disasters(value=disasters_array, params_of_mean=params_of_mean):
    """Annual occurences of coal mining disasters."""
    val = params_of_mean[1] + params_of_mean[0] * arange(111)
    return poisson_like(value, val)



__docformat__ = 'reStructuredText'
__author__ = 'Anand Patil, anand.prabhakar.patil@gmail.com'
__all__ = [
    'extend_children',
    'extend_parents',
    'ParentDict',
    'Stochastic',
    'Deterministic',
    'Potential']


from copy import copy
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
from numpy import array, ndarray, reshape, Inf, asarray, dot, sum, float, isnan, size, NaN, asanyarray
import numpy as np
from numpy import shape, size, ravel, zeros, ones, reshape, newaxis, broadcast, ndim, expand_dims
from .Node import Node, ZeroProbability, Variable, PotentialBase, StochasticBase, DeterministicBase
from . import Container
from .Container import DictContainer, ContainerBase, file_items, ArrayContainer
import sys
import pdb
from . import calc_utils

from . import datatypes

from . import six
from .six import print_

d_neg_inf = float(-1.7976931348623157e+308)

# from PyrexLazyFunction import LazyFunction
from .LazyFunction import LazyFunction, Counter


def extend_children(children):
    """
    extend_children(children)

    Returns a set containing
    nearest conditionally stochastic (Stochastic, not Deterministic) descendants.
    """
    new_children = copy(children)
    need_recursion = False
    dtrm_children = set()

    for child in children:
        if isinstance(child, Deterministic):
            new_children |= child.children
            dtrm_children.add(child)
            need_recursion = True

    new_children -= dtrm_children

    if need_recursion:
        new_children = extend_children(new_children)

    return new_children


def extend_parents(parents):
    """
    extend_parents(parents)

    Returns a set containing
    nearest conditionally stochastic (Stochastic, not Deterministic) ancestors.
    """
    new_parents = set()

    for parent in parents:

        new_parents.add(parent)

        if isinstance(parent, DeterministicBase):
            new_parents.remove(parent)
            new_parents |= parent.extended_parents

        elif isinstance(parent, ContainerBase):
            for contained_parent in parent.stochastics:
                new_parents.add(contained_parent)
            for contained_parent in parent.deterministics:
                new_parents |= contained_parent.extended_parents

    return new_parents


class ParentDict(DictContainer):

    """
    A special subclass of DictContainer which makes it safe to change
    variables' parents. When __setitem__ is called, a ParentDict instance
    removes its owner from the old parent's children set (if appropriate)
    and adds its owner to the new parent's children set. It then asks
    its owner to generate a new LazyFunction instance using its new
    parents.

    Also manages the extended_parents attribute of owner.

    NB: StepMethod and Model are expecting variables'
    children to be static. If you want to change independence structure
    over the course of an MCMC loop, please do so with indicator variables.

    :SeeAlso: DictContainer
    """

    def __init__(self, regular_dict, owner):
        DictContainer.__init__(self, dict(regular_dict))
        self.owner = owner
        self.owner.extended_parents = extend_parents(self.variables)
        if isinstance(self.owner, StochasticBase) or isinstance(
                self.owner, PotentialBase):
            self.has_logp = True
        else:
            self.has_logp = False

    def detach_parents(self):
        for parent in six.itervalues(self):
            if isinstance(parent, Variable):
                parent.children.discard(self.owner)
            elif isinstance(parent, ContainerBase):
                for variable in parent.variables:
                    variable.chidren.discard(self.owner)

        if self.has_logp:
            self.detach_extended_parents()

    def detach_extended_parents(self):
        for e_parent in self.owner.extended_parents:
            if isinstance(e_parent, StochasticBase):
                e_parent.extended_children.discard(self.owner)

    def attach_parents(self):
        for parent in six.itervalues(self):
            if isinstance(parent, Variable):
                parent.children.add(self.owner)
            elif isinstance(parent, ContainerBase):
                for variable in parent.variables:
                    variable.children.add(self.owner)

        if self.has_logp:
            self.attach_extended_parents()

    def attach_extended_parents(self):
        for e_parent in self.owner.extended_parents:
            if isinstance(e_parent, StochasticBase):
                e_parent.extended_children.add(self.owner)

    def __setitem__(self, key, new_parent):
        old_parent = self[key]

        # Possibly remove owner from old parent's children set.
        if isinstance(old_parent, Variable) or isinstance(
                old_parent, ContainerBase):

            # Tell all extended parents to forget about owner
            if self.has_logp:
                self.detach_extended_parents()

            self.val_keys.remove(key)
            self.nonval_keys.append(key)

            if isinstance(old_parent, Variable):
                # See if owner only claims the old parent via this key.
                if sum(
                        [parent is old_parent for parent in six.itervalues(self)]) == 1:
                    old_parent.children.remove(self.owner)

            if isinstance(old_parent, ContainerBase):
                for variable in old_parent.variables:
                    if sum(
                            [parent is variable for parent in six.itervalues(self)]) == 1:
                        variable.children.remove(self.owner)

        # If the new parent is a variable, add owner to its children set.
        if isinstance(new_parent, Variable) or isinstance(
                new_parent, ContainerBase):

            self.val_keys.append(key)
            self.nonval_keys.remove(key)

            if isinstance(new_parent, Variable):
                new_parent.children.add(self.owner)

            elif isinstance(new_parent, ContainerBase):
                for variable in new_parent.variables:
                    new_parent.children.add(self.owner)

        # Totally recompute extended parents
        self.owner.extended_parents = extend_parents(self.variables)
        if self.has_logp:
            self.attach_extended_parents()

        dict.__setitem__(self, key, new_parent)

        file_items(self, self)

        # Tell my owner it needs a new lazy function.
        self.owner.gen_lazy_function()


class Potential(PotentialBase):

    """
    Not a variable; just an arbitrary log-probability term to multiply into the
    joint distribution. Useful for expressing models that aren't directed, such as
    Markov random fields.

    Decorator instantiation:

    @potential(trace = True)
    def A(x = B, y = C):
        return -.5 * (x-y)**2 / 3.

    Direct instantiation:

    :Parameters:

        -logp: function
              The function that computes the potential's value from the values
              of its parents.

        -doc: string
              The docstring for this potential.

        -name: string
              The name of this potential.

        -parents: dictionary
              A dictionary containing the parents of this potential.

        -cache_depth (optional): integer
              An integer indicating how many of this potential's value computations
              should be 'memoized'.

        - plot (optional) : boolean
            A flag indicating whether this variable is to be plotted.

        - verbose (optional) : integer
              Level of output verbosity: 0=none, 1=low, 2=medium, 3=high


    Externally-accessible attribute:

        -logp: float
              Returns the potential's log-probability given its parents' values. Skips
              computation if possible.

    No methods.

    :SeeAlso: Stochastic, Node, LazyFunction, stoch, dtrm, data, Model, Container
    """

    def __init__(self, logp, doc, name, parents, cache_depth=2,
                 plot=None, verbose=-1, logp_partial_gradients=None):

        if logp_partial_gradients is None:
            logp_partial_gradients = {}

        self.ParentDict = ParentDict

        # This function gets used to evaluate self's value.
        self._logp_fun = logp
        self._logp_partial_gradients_functions = logp_partial_gradients

        self.errmsg = "Potential %s forbids its parents' current values" % name

        Node.__init__(self,
                      doc=doc,
                      name=name,
                      parents=parents,
                      cache_depth=cache_depth,
                      verbose=verbose)

        self._plot = plot

        # self._logp.force_compute()

        # Check initial value
        if not isinstance(self.logp, float):
            raise ValueError(
                "Potential " +
                self.__name__ +
                "'s initial log-probability is %s, should be a float." %
                self.logp.__repr__(
                ))

    def gen_lazy_function(self):

        self._logp = LazyFunction(fun=self._logp_fun,
                                  arguments=self.parents,
                                  ultimate_args=self.extended_parents,
                                  cache_depth=self._cache_depth)
        self._logp.force_compute()

        self._logp_partial_gradients = {}
        for parameter, function in six.iteritems(self._logp_partial_gradients_functions):
            lazy_logp_partial_gradients = LazyFunction(fun=function,
                                                       arguments=self.parents,
                                                       ultimate_args=self.extended_parents,
                                                       cache_depth=self._cache_depth)
            lazy_logp_partial_gradients.force_compute()
            self._logp_partial_gradients[
                parameter] = lazy_logp_partial_gradients

    def get_logp(self):

        if self.verbose > 1:
            print_('\t' + self.__name__ + ': log-probability accessed.')
        logp = self._logp.get()
        if self.verbose > 1:
            print_('\t' + self.__name__ + ': Returning log-probability ', logp)

        try:
            logp = float(logp)
        except:
            raise TypeError(
                self.__name__ + ': computed log-probability ' + str(
                    logp) + ' cannot be cast to float')

        if logp != logp:
            raise ValueError(
                self.__name__ +
                ': computed log-probability is NaN')

        # Check if the value is smaller than a double precision infinity:
        if logp <= d_neg_inf:
            if self.verbose > 0:
                raise ZeroProbability(
                    self.errmsg + ": %s" %
                    self._parents.value)
            else:
                raise ZeroProbability(self.errmsg)

        return logp

    def set_logp(self, value):
        raise AttributeError(
            'Potential ' +
            self.__name__ +
            '\'s log-probability cannot be set.')

    logp = property(
        fget=get_logp,
        fset=set_logp,
        doc="Self's log-probability value conditional on parents.")

    def logp_partial_gradient(self, variable, calculation_set=None):
        gradient = 0
        if self in calculation_set:

            if not datatypes.is_continuous(variable):
                return zeros(shape(variable.value))

            for parameter, value in six.iteritems(self.parents):

                if value is variable:
                    try:
                        grad_func = self._logp_partial_gradients[parameter]
                    except KeyError:
                        raise NotImplementedError(
                            repr(
                                self) +
                            " has no gradient function for parameter " +
                            parameter)

                    gradient = gradient + grad_func.get()

        # np.reshape(gradient, np.shape(variable.value))
        return np.reshape(gradient, np.shape(variable.value))


class Deterministic(DeterministicBase):

    """
    A variable whose value is determined by the values of its parents.

    Decorator instantiation:

    @dtrm(trace=True)
    def A(x = B, y = C):
        return sqrt(x ** 2 + y ** 2)

    :Parameters:
      eval : function
        The function that computes the variable's value from the values
        of its parents.
      doc : string
        The docstring for this variable.
      name: string
        The name of this variable.
      parents: dictionary
        A dictionary containing the parents of this variable.
      trace (optional): boolean
        A boolean indicating whether this variable's value
        should be traced (in MCMC).
      cache_depth (optional): integer
        An integer indicating how many of this variable's
        value computations should be 'memoized'.
      plot (optional) : boolean
        A flag indicating whether this variable is to be plotted.
      verbose (optional) : integer
        Level of output verbosity: 0=none, 1=low, 2=medium, 3=high
      jacobian (optional) : function(parameter, **same args as function)
        function which calculates the analytical jacobian for the deterministic with respect to some parameter
      jacobian_format (optional) : dict <string, string>
        formats of the jacobians returned by the jacobian function for each parameter:
            'full' : the function returns the full jacobian
            'broadcast_operation' : the function returns the jacobian for an operation where the argument arrays are broadcast to eachother
            'accumulation_operation' : the function returns the jacobian for an operation where the number of dimensions is reduced
        the default is 'full'

    :Attributes:
      value : any object
        Returns the variable's value given its parents' values. Skips
        computation if possible.

    :SeeAlso:
      Stochastic, Potential, deterministic, MCMC, Lambda,
      LinearCombination, Index
    """
    __array_priority__ = 1000

    def __init__(self, eval, doc, name, parents, dtype=None, trace=True,
                 cache_depth=2, plot=None, verbose=-1, jacobians={}, jacobian_formats={}):

        self.ParentDict = ParentDict

        # This function gets used to evaluate self's value.
        self._eval_fun = eval
        self._jacobian_functions = jacobians
        self._jacobian_formats = jacobian_formats

        Variable.__init__(self,
                          doc=doc,
                          name=name,
                          parents=parents,
                          cache_depth=cache_depth,
                          dtype=dtype,
                          trace=trace,
                          plot=plot,
                          verbose=verbose)

        # self._value.force_compute()

    def gen_lazy_function(self):

        self._value = LazyFunction(fun=self._eval_fun,
                                   arguments=self.parents,
                                   ultimate_args=self.extended_parents,
                                   cache_depth=self._cache_depth)

        self._value.force_compute()

        self._jacobians = {}
        for parameter, function in six.iteritems(self._jacobian_functions):
            lazy_jacobian = LazyFunction(fun=function,
                                         arguments=self.parents,
                                         ultimate_args=self.extended_parents,
                                         cache_depth=self._cache_depth)
            lazy_jacobian.force_compute()
            self._jacobians[parameter] = lazy_jacobian

    def get_value(self):
        if self.verbose > 1:
            print_('\t' + self.__name__ + ': value accessed.')
        _value = self._value.get()
        if isinstance(_value, ndarray):
            _value.flags['W'] = False
        if self.verbose > 1:
            print_('\t' + self.__name__ + ': Returning value ', _value)
        return _value

    def set_value(self, value):
        raise AttributeError(
            'Deterministic ' +
            self.__name__ +
            '\'s value cannot be set.')

    value = property(
        fget=get_value,
        fset=set_value,
        doc="Self's value computed from current values of parents.")

    def apply_jacobian(self, parameter, variable, gradient):
        try:
            jacobian_func = self._jacobians[parameter]
        except KeyError:
            raise NotImplementedError(
                repr(
                    self) +
                " has no jacobian function for parameter " +
                parameter)

        jacobian = jacobian_func.get()

        mapping = self._jacobian_formats.get(parameter, 'full')

        p = self._format_mapping[mapping](self, variable, jacobian, gradient)
        return p

    def logp_partial_gradient(self, variable, calculation_set=None):
        """
        gets the logp gradient of this deterministic with respect to variable
        """
        if self.verbose > 0:
            print_('\t' + self.__name__ + ': logp_partial_gradient accessed.')

        if not (datatypes.is_continuous(variable)
                and datatypes.is_continuous(self)):
            return zeros(shape(variable.value))

        # loop through all the parameters and add up all the gradients of log p
        # with respect to the approrpiate variable
        gradient = builtins.sum(
            [child.logp_partial_gradient(self,
                                         calculation_set) for child in self.children])

        totalGradient = 0
        for parameter, value in six.iteritems(self.parents):
            if value is variable:

                totalGradient += self.apply_jacobian(
                    parameter, variable, gradient)

        return np.reshape(totalGradient, shape(variable.value))

    def full_jacobian(self, variable, jacobian, gradient):
        return dot(np.transpose(jacobian), np.ravel(gradient)[:, np.newaxis])

    def transformation_operation_jacobian(self, variable, jacobian, gradient):
        return jacobian * gradient

    def broadcast_operation_jacobian(self, variable, jacobian, gradient):
        return calc_utils.sum_to_shape(
            id(variable), id(self), jacobian * gradient, shape(variable.value))

    def accumulation_operation_jacobian(self, variable, jacobian, gradient):
        for i in range(ndim(jacobian)):
            if i >= ndim(gradient) or shape(
                    gradient)[i] != shape(variable.value)[i]:
                expand_dims(gradient, i)
        return gradient * jacobian

    def index_operation_jacobian(self, variable, jacobian, gradient):
        derivative = zeros(shape(variable.value))
        derivative[jacobian] = gradient
        return derivative

    _format_mapping = {'full': full_jacobian,
                       'transformation_operation':
                       transformation_operation_jacobian,
                       'broadcast_operation': broadcast_operation_jacobian,
                       'accumulation_operation':
                       accumulation_operation_jacobian,
                       'index_operation': index_operation_jacobian}


class Stochastic(StochasticBase):

    """
    A variable whose value is not determined by the values of its parents.


    Decorator instantiation:

    @stoch(trace=True)
    def X(value = 0., mu = B, tau = C):
        return Normal_like(value, mu, tau)

    @stoch(trace=True)
    def X(value=0., mu=B, tau=C):

        def logp(value, mu, tau):
            return Normal_like(value, mu, tau)

        def random(mu, tau):
            return Normal_r(mu, tau)

        rseed = 1.


    Direct instantiation:



    - logp : function
            The function that computes the variable's log-probability from
            its value and the values of its parents.

    - doc : string
            The docstring for this variable.

    - name : string
            The name of this variable.

    - parents: dict
            A dictionary containing the parents of this variable.

    - random (optional) : function
            A function that draws a new value for this
            variable given its parents' values.

    - trace (optional) : boolean
            A boolean indicating whether this variable's value
            should be traced (in MCMC).

    - value (optional) : number or array
            An initial value for this variable

    - dtype (optional) : type
            A type for this variable.

    - rseed (optional) : integer or rseed
            A seed for this variable's rng. Either value or rseed must
            be given.

    - observed (optional) :  boolean
            A flag indicating whether this variable is data; whether
            its value is known.

    - cache_depth (optional) : integer
            An integer indicating how many of this variable's
            log-probability computations should be 'memoized'.

    - plot (optional) : boolean
            A flag indicating whether this variable is to be plotted.

    - verbose (optional) : integer
            Level of output verbosity: 0=none, 1=low, 2=medium, 3=high


    Externally-accessible attribute:

    - value: any class
          Returns this variable's current value.

    - logp: float
          Returns the variable's log-probability given its value and its
          parents' values. Skips computation if possible.

    last_value: any class
          Returns this variable's last value. Useful for rejecting
          Metropolis-Hastings jumps. See touch() and the warning below.

    Externally-accessible methods:

    random():   Draws a new value for this variable from its distribution and
                returns it.

    :SeeAlso: Deterministic, Node, LazyFunction, stoch, dtrm, data, Model, Container
    """
    __array_priority__ = 1000

    def __init__(self,
                 logp,
                 doc,
                 name,
                 parents,
                 random=None,
                 trace=True,
                 value=None,
                 dtype=None,
                 rseed=False,
                 observed=False,
                 cache_depth=2,
                 plot=None,
                 verbose=-1,
                 check_logp=True,
                 logp_partial_gradients=None):

        if logp_partial_gradients is None:
            logp_partial_gradients = {}

        self.counter = Counter()
        self.ParentDict = ParentDict

        # A flag indicating whether self's value has been observed.
        self._observed = observed
        # Default value of None for mask
        self._mask = None
        if observed:

            if value is None:
                raise ValueError(
                    'Stochastic %s must be given an initial value if observed=True.' %
                    name)

            try:

                # If there are missing values, store mask to missing elements
                self._mask = value.mask

                # This check ensures that the mask has missing values
                if value.mask.sum():

                    # Set to value of mean of observed data
                    if value.fill_value == '?':
                        value.fill_value = value.mean()
                    value = value.filled()

                    # Set observed flag to False, so that missing values will
                    # update
                    self._observed = False

            except AttributeError:
                # Must not have missing values
                pass

        # This function will be used to evaluate self's log probability.
        self._logp_fun = logp

        # This function will be used to evaluate self's gradient of log
        # probability.
        self._logp_partial_gradient_functions = logp_partial_gradients

        # This function will be used to draw values for self conditional on
        # self's parents.
        self._random = random

        # A seed for self's rng. If provided, the initial value will be drawn. Otherwise it's
        # taken from the constructor.
        self.rseed = rseed

        self.errmsg = "Stochastic %s's value is outside its support,\n or it forbids its parents' current values." % name

        dtype = np.dtype(dtype)

        # Initialize value, either from value provided or from random function.
        try:
            if dtype.kind != 'O' and value is not None:
                # Convert Pandas DataFrames and Series to numpy arrays
                value = getattr(value, 'values', value)
                self._value = np.array(value, dtype=dtype)
                self._value.flags['W'] = False
            else:
                self._value = value
        except:
            cls, inst, tb = sys.exc_info()
            new_inst = cls(
                'Stochastic %s: Failed to cast initial value to required dtype.\n\nOriginal error message:\n' %
                name + inst.message)
            six.reraise(cls, new_inst, tb)

        # Store the shape of the stochastic value
        self._shape = np.shape(self._value)

        Variable.__init__(self,
                          doc=doc,
                          name=name,
                          parents=parents,
                          cache_depth=cache_depth,
                          trace=trace,
                          dtype=dtype,
                          plot=plot,
                          verbose=verbose)

        # self._logp.force_compute()

        self._shape = np.shape(self._value)

        if isinstance(self._value, ndarray):
            self._value.flags['W'] = False

        if check_logp:
            # Check initial value
            if not isinstance(self.logp, float):
                raise ValueError(
                    "Stochastic " +
                    self.__name__ +
                    "'s initial log-probability is %s, should be a float." %
                    self.logp.__repr__(
                    ))

    def gen_lazy_function(self):
        """
        Will be called by Node at instantiation.
        """

        # If value argument to __init__ was None, draw value from random
        # method.
        if self._value is None:

            # Use random function if provided
            if self._random is not None:
                self.value = self._random(**self._parents.value)

            # Otherwise leave initial value at None and warn.
            else:
                raise ValueError(
                    'Stochastic ' +
                    self.__name__ +
                    "'s value initialized to None; no initial value or random method provided.")

        arguments = {}
        arguments.update(self.parents)
        arguments['value'] = self
        arguments = DictContainer(arguments)

        self._logp = LazyFunction(fun=self._logp_fun,
                                  arguments=arguments,
                                  ultimate_args=self.extended_parents | set(
                                      [self]),
                                  cache_depth=self._cache_depth)
        self._logp.force_compute()

        self._logp_partial_gradients = {}

        for parameter, function in six.iteritems(self._logp_partial_gradient_functions):
            lazy_logp_partial_gradient = LazyFunction(fun=function,
                                                      arguments=arguments,
                                                      ultimate_args=self.extended_parents | set(
                                                          [self]),
                                                      cache_depth=self._cache_depth)
            # lazy_logp_partial_gradient.force_compute()
            self._logp_partial_gradients[parameter] = lazy_logp_partial_gradient

    def get_value(self):
        # Define value attribute
        if self.verbose > 1:
            print_('\t' + self.__name__ + ': value accessed.')
        return self._value

    def get_stoch_value(self):
        if self.verbose > 1:
            print_('\t' + self.__name__ + ': stoch_value accessed.')
        return self._value[self.mask]

    def set_value(self, value, force=False):
        # Record new value and increment counter

        # Value can't be updated if observed=True
        if self.observed and not force:
            raise AttributeError(
                'Stochastic ' +
                self.__name__ +
                '\'s value cannot be updated if observed flag is set')

        if self.verbose > 0:
            print_('\t' + self.__name__ + ': value set to ', value)

        # Save current value as last_value
        # Don't copy because caching depends on the object's reference.
        self.last_value = self._value

        if self.mask is None:

            if self.dtype.kind != 'O':
                self._value = asanyarray(value, dtype=self.dtype)
                self._value.flags['W'] = False
            else:
                self._value = value

        else:

            new_value = self.value.copy()

            new_value[
                self.mask] = asanyarray(
                    value,
                    dtype=self.dtype)[
                        self.mask]
            self._value = new_value

        self.counter.click()

    value = property(
        fget=get_value,
        fset=set_value,
        doc="Self's current value.")

    def mask():
        doc = "Returns the mask for missing values"

        def fget(self):
            return self._mask
        return locals()
    mask = property(**mask())

    def shape():
        doc = "The shape of the value of self."

        def fget(self):
            if self.verbose > 1:
                print_('\t' + self.__name__ + ': shape accessed.')
            return self._shape
        return locals()
    shape = property(**shape())

    def revert(self):
        """
        Sets self's value to self's last value. Bypasses the data cleaning in
        the set_value method.
        """
        self.counter.unclick()
        self._value = self.last_value

    def get_logp(self):

        if self.verbose > 0:
            print_('\t' + self.__name__ + ': logp accessed.')
        logp = self._logp.get()

        try:
            logp = float(logp)
        except:
            raise TypeError(
                self.__name__ + ': computed log-probability ' + str(
                    logp) + ' cannot be cast to float')

        if logp != logp:
            return -np.inf

        if self.verbose > 0:
            print_('\t' + self.__name__ + ': Returning log-probability ', logp)

        # Check if the value is smaller than a double precision infinity:
        if logp <= d_neg_inf:
            if self.verbose > 0:
                raise ZeroProbability(
                    self.errmsg + "\nValue: %s\nParents' values:%s" %
                    (self._value, self._parents.value))
            else:
                raise ZeroProbability(self.errmsg)

        return logp

    def set_logp(self, new_logp):
        raise AttributeError(
            'Stochastic ' +
            self.__name__ +
            '\'s logp attribute cannot be set')

    logp = property(
        fget=get_logp,
        fset=set_logp,
        doc="Log-probability or log-density of self's current value\n given values of parents.")

    def logp_gradient_contribution(self, calculation_set=None):
        """
        Calculates the gradient of the joint log posterior with respect to self.
        Calculation of the log posterior is restricted to the variables in calculation_set.
        """
        # NEED some sort of check to see if the log p calculation has recently
        # failed, in which case not to continue

        return self.logp_partial_gradient(self, calculation_set) + builtins.sum(
            [child.logp_partial_gradient(self, calculation_set) for child in self.children])

    def logp_partial_gradient(self, variable, calculation_set=None):
        """
        Calculates the partial gradient of the posterior of self with respect to variable.
        Returns zero if self is not in calculation_set.
        """
        if (calculation_set is None) or (self in calculation_set):

            if not datatypes.is_continuous(variable):
                return zeros(shape(variable.value))

            if variable is self:
                try:
                    gradient_func = self._logp_partial_gradients['value']

                except KeyError:
                    raise NotImplementedError(
                        repr(
                            self) +
                        " has no gradient function for 'value'")

                gradient = np.reshape(
                    gradient_func.get(
                    ),
                    np.shape(
                        variable.value))
            else:
                gradient = builtins.sum(
                    [self._pgradient(variable,
                                     parameter,
                                     value) for parameter,
                     value in six.iteritems(self.parents)])

            return gradient
        else:
            return 0

    def _pgradient(self, variable, parameter, value):
        if value is variable:
            try:
                return np.reshape(
                    self._logp_partial_gradients[parameter].get(), np.shape(variable.value))
            except KeyError:
                raise NotImplementedError(
                    repr(
                        self) +
                    " has no gradient function for parameter " +
                    parameter)
        else:
            return 0

    # Sample self's value conditional on parents.
    def random(self):
        """
        Draws a new value for a stoch conditional on its parents
        and returns it.

        Raises an error if no 'random' argument was passed to __init__.
        """

        if self._random:
            # Get current values of parents for use as arguments for _random()
            r = self._random(**self.parents.value)
        else:
            raise AttributeError(
                'Stochastic ' +
                self.__name__ +
                ' does not know how to draw its value, see documentation')

        if self.shape:
            r = np.reshape(r, self.shape)

        # Set Stochastic's value to drawn value
        if not self.observed:
            self.value = r

        return r

    # Shortcut alias to random
    rand = random

    def _get_observed(self):
        return self._observed

    def _set_observed(self, observed):
        raise ValueError(
            'Stochastic %s: "observed" flag cannot be changed.' %
            self.__name__)
    observed = property(_get_observed, _set_observed)

    def _get_coparents(self):
        coparents = set()
        for child in self.extended_children:
            coparents |= child.extended_parents
        coparents.add(self)
        return coparents
    coparents = property(
        _get_coparents,
        doc="All the variables whose extended children intersect with self's.")

    def _get_moral_neighbors(self):
        moral_neighbors = self.coparents | self.extended_parents | self.extended_children
        for neighbor in copy(moral_neighbors):
            if isinstance(neighbor, PotentialBase):
                moral_neighbors.remove(neighbor)
        return moral_neighbors
    moral_neighbors = property(
        _get_moral_neighbors,
        doc="Self's neighbors in the moral graph: self's Markov blanket with self removed.")

    def _get_markov_blanket(self):
        return self.moral_neighbors | set([self])
    markov_blanket = property(
        _get_markov_blanket,
        doc="Self's coparents, self's extended parents, self's children and self.")



import pymc
from pymc.examples import gelman_bioassay
import numpy as np
from numpy.testing import *

# Easy dice example (analytically solvable)


def dice(data=None):

    if data is None:
        x = [pymc.rbernoulli(1.0 / 6.0) for i in range(0, 100)]
    else:
        x = data

    prob = pymc.Uniform('prob', lower=0, upper=1)

    d = pymc.Bernoulli('d', p=prob, value=x, observed=True)

    return locals()


class TestSlice(TestCase):

    def test_dice_sample(self):
        M = pymc.MCMC(dice())
        M.use_step_method(pymc.Slicer, M.prob, w=.1)
        M.sample(iter=1000, tune_interval=1)

    def test_bioassay_sample(self):
        M = pymc.MCMC(gelman_bioassay)
        for stoch in M.stochastics:
            M.use_step_method(pymc.Slicer, stoch, w=.1)
        M.sample(iter=1000, tune_interval=1)



from numpy.testing import *
from pymc.gp import *
from pymc.gp.cov_funs import *
from numpy import *
from copy import copy

from pymc import six
xrange = six.moves.xrange

random.seed(42)

def zero_fun(x):
    return zeros(x.shape[:-1], dtype=float)
M = Mean(zero_fun)

lin = linspace(-1., 1., 50)
x, y = meshgrid(lin, lin)
xstack = zeros(x.shape + (2,), dtype=float)
xstack[:, :, 0] = x
xstack[:, :, 1] = y

lin_obs = linspace(-.5, .5, 2)
xobs, yobs = meshgrid(lin_obs, lin_obs)
ostack = zeros(xobs.shape + (2,), dtype=float)
ostack[:, :, 0] = xobs
ostack[:, :, 1] = yobs

V = .02 * ones((2, 2), dtype=float)
data = ones((2, 2), dtype=float)
N = 10

coef_cov = ones((2 * N + 1, 2 * N + 1), dtype=float)
for i in xrange(1, 2 * N + 1):
    int_i = int(((i + 1) / 2))
    for j in xrange(1, 2 * N + 1):
        int_j = int(((j + 1) / 2))
        coef_cov[i, j] = 1. / (sqrt((int_i + int_j) ** 2)) ** 2.5

basis = fourier_basis([N, N])
C = SeparableBasisCovariance(basis, coef_cov, xmin=[-4., -.4], xmax=[4., 4.])


class test_basiscov(TestCase):

    def test_sep(self):

        observe(M, C, obs_mesh=ostack, obs_V=V, obs_vals=data)

    def test_nonsep(self):

        basis_array = zeros((2 * N + 1, 2 * N + 1), dtype=object)
        for i in arange(2 * N + 1):

            if i % 2 == 0:
                def funi(x, xmin, xmax, i=i):
                    T = xmax[0] - xmin[0]
                    return cos(i * pi * (x - xmin[0]) / T)

            else:
                def funi(x, xmin, xmax, i=i):
                    T = xmax[0] - xmin[0]
                    return sin((i + 1) * pi * (x - xmin[0]) / T)

            for j in arange(2 * N + 1):

                if j % 2 == 0:
                    def funj(x, xmin, xmax, j=j):
                        T = xmax[1] - xmin[1]
                        return cos(j * pi * (x - xmin[1]) / T)
                else:
                    def funj(x, xmin, xmax, j=j):
                        T = xmax[1] - xmin[1]
                        return sin((j + 1) * pi * (x - xmin[1]) / T)

                def fun_now(x, xmin, xmax, funi=funi, funj=funj):
                    return funi(x[:, 0], xmin, xmax) * \
                        funj(x[:, 1], xmin, xmax)

                basis_array[i, j] = fun_now

        C2 = BasisCovariance(
            basis_array, coef_cov, xmin=[-4., -.4], xmax=[4., 4.])
        assert_almost_equal(C2(xstack), C(xstack))

        observe(M, C, obs_mesh=ostack, obs_V=V, obs_vals=data)



from numpy.testing import *
from numpy import *
from pymc.examples import disaster_model as DM
from pymc import Container, Normal


class test_Container(TestCase):

    def test_container_parents(self):
        A = Normal('A', 0, 1)
        B = Normal('B', 0, 1)

        C = Normal('C', [A, B], 1)

        assert_equal(Container([A, B]).value, [A.value, B.value])
        assert_equal(C.parents.value['mu'], [A.value, B.value])

    def test_nested_tuple_container(self):
        A = Normal('A', 0, 1)
        try:
            Container(([A],))
            raise AssertionError('A NotImplementedError should have resulted.')
        except NotImplementedError:
            pass

    def test(self):

        # Test set container:

        S = [DM.early_mean, DM.switchpoint, DM.late_mean]
        R = Container(S)

        for item in R:
            assert(item in S)

        for item in R.value:
            val_in_S = False
            for S_item in S:
                if S_item.value is item:
                    val_in_S = True

            assert(val_in_S)

        # Test list/dictionary container:

        A = [[DM.early_mean, DM.switchpoint],
             [DM.late_mean, DM.disasters, 3.], 54.323]
        C = Container(A)

        for i in range(2):
            for j in range(2):
                assert(C[i][j] == A[i][j])
                assert(all(C.value[i][j] == A[i][j].value))

            assert(C[1][2] == A[1][2])
            assert(C.value[1][2] == A[1][2])

            assert(C[2] == A[2])
            assert(C.value[2] == A[2])

        # Test array container:

        B = ndarray((3, 3), dtype=object)
        B[0, :] = DM.early_mean
        B[1, :] = 1.
        B[2, :] = A
        D = Container(B)

        for i in range(2):
            assert(D[0, i] == DM.early_mean)
            assert(D.value[0, i] == DM.early_mean.value)
            assert(D[1, i] == 1.)
            assert(D.value[1, i] == 1.)

        P = D[2, :]
        Q = D.value[2, :]

        for i in range(2):
            for j in range(2):
                assert(P[i][j] == A[i][j])
                assert(all(Q[i][j] == A[i][j].value))

            assert(P[1][2] == A[1][2])
            assert(Q[1][2] == A[1][2])

            assert(P[2] == A[2])
            assert(Q[2] == A[2])


if __name__ == '__main__':
    import unittest
    unittest.main()



print("import: 'pymc'")
import pymc

print("import: 'pymc.database'")
import pymc.database

print("import: 'pymc.examples'")
import pymc.examples

print("import: 'pymc.gp'")
import pymc.gp

print("import: 'pymc.gp.cov_funs'")
import pymc.gp.cov_funs

print("import: 'pymc.Container_values'")
import pymc.Container_values

print("import: 'pymc.LazyFunction'")
import pymc.LazyFunction

print("import: 'pymc.flib'")
import pymc.flib

print("import: 'pymc.gp.cov_funs.distances'")
import pymc.gp.cov_funs.distances

print("import: 'pymc.gp.cov_funs.isotropic_cov_funs'")
import pymc.gp.cov_funs.isotropic_cov_funs

print("import: 'pymc.gp.incomplete_chol'")
import pymc.gp.incomplete_chol

print("import: 'pymc.gp.linalg_utils'")
import pymc.gp.linalg_utils




from . import PyMCmodel
import pymc as pm
from pymc import six
import numpy as np
import matplotlib.pyplot as pl
import matplotlib
matplotlib.rcParams['axes.facecolor'] = 'w'
import time

n_iter = 10000

rej = []
times = []
mesh_sizes = [0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512]
for mesh_size in [0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512]:
    six.print_(mesh_size)
    m = PyMCmodel.make_model(mesh_size, False)
    if mesh_size == 0:
        sm = pm.gp.MeshlessGPMetropolis(m['sm'].f)
    else:
        sm = pm.gp.GPEvaluationMetropolis(m['sm'].f_eval, proposal_sd=.01)
    t1 = time.time()
    for i in xrange(n_iter):
        sm.step()
    times.append((time.time() - t1) / float(n_iter))
    rej.append(sm.rejected / float(sm.rejected + sm.accepted))

m = PyMCmodel.make_model(0, True)
sm = pm.gp.GPEvaluationMetropolis(m['sm'].f_eval, proposal_sd=.01)
t1 = time.time()
for i in xrange(n_iter):
    sm.step()
truemesh_time = (time.time() - t1) / float(n_iter)
truemesh_rej = sm.rejected / float(sm.rejected + sm.accepted)


f = pl.figure(1, figsize=(12, 6))
ax1 = f.add_subplot(111)
ax1.plot(rej, 'b-', label='Rejection', linewidth=4)
ax1.plot([3], [truemesh_rej], 'b.', markersize=8)
ax1.set_ylabel('Rejection rate', color='b')
ax1.set_xticks(range(len(mesh_sizes)))
ax1.set_xticklabels(mesh_sizes)
ax1.set_xlabel('Number of points in mesh')


ax2 = ax1.twinx()
ax2.plot(times, 'r-', label='Time', linewidth=4)
ax2.plot([3], [truemesh_time], 'r.', markersize=8)
ax2.set_ylabel('Time per jump (seconds)', color='r')
ax2.set_xticks(range(len(mesh_sizes)))
ax2.set_xticklabels(mesh_sizes)
ax2.set_xlabel('Number of points in mesh')



from .isotropic_cov_funs import *
from .cov_utils import *
from .bases import *
from .wrapped_distances import *
from . import isotropic_cov_funs
from .brownian import *
from .nsmatern import *



extra_parameters = {'gaussian': {'': ''},
                'pow_exp': {'pow': 'The exponent in the exponential.'},
                'exponential':{'':''},
                'matern': {'diff_degree': 'The degree of differentiability of realizations.'},
                'sphere': {'': ''},
                'quadratic': {'phi': 'The characteristic (scaled) distance of decorrelation.'},
                'exponential': {'': ''}}


for name in extra_parameters:
    locals()[name] = covariance_function_bundle(name, 'isotropic_cov_funs', extra_parameters[name], ampsq_is_diag=True)

nsmatern_extra_params = {'diff_degree': 'A function giving the local degree of differentiability.',
                         'h': 'A function giving the local relative amplitude.'}
nsmatern = covariance_function_bundle('nsmatern', 'nsmatern',
                        nsmatern_extra_params, ampsq_is_diag=False, with_x=True)
for w in nsmatern.wrappers:
    w.diag_call = nsmatern_diag




# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'
__all__ = ['FullRankCovariance']


from numpy import *
from numpy.linalg import cholesky, LinAlgError
from .GPutils import regularize_array, trisolve
from .linalg_utils import dpotrf_wrap
from .Covariance import Covariance
from .incomplete_chol import ichol, ichol_continue

from pymc import six
xrange = six.moves.xrange


class FullRankCovariance(Covariance):

    """
    C=FullRankCovariance(eval_fun, **params)

    A GP covariance.

    All linear algebra done with dense BLAS, so attempts to invert/ factorize
    numerically singular covariance matrices will cause errors. On the other
    hand, computations will be faster than with Covariance for full-rank
    covariance matrices.

    :Arguments:

        -   `eval_fun`: A function that takes either a single value x or two values x and y,
            followed by an arbitrary number of keyword parameters. x and y will be of shape
            (n,n_dim), where n is any integer and n_dim is the dimensionality of the space, or
            shape (n). In the latter case n_dim should be assumed to be 1.

        -   `params`: Parameters to be passed to eval_fun.


    :SeeAlso: Mean, BasisCovariance, SeparableBasisCovariance, Realization, observe
    """

    def __init__(self, eval_fun, nugget=None, **params):

        self.ndim = None
        self.observed = False
        self.obs_mesh = None
        self.obs_V = None
        self.Uo = None
        self.obs_piv = None
        self.obs_len = None
        self.full_piv = None
        self.full_obs_mesh = None
        self.basiscov = False

        self.eval_fun = eval_fun
        self.params = params
        self.nugget = nugget

        # Sorry... the diagonal calls are done using f2py for speed.
        # def diag_cov_fun(xe):
        #     return self.eval_fun(xe,xe,**self.params)
        #
        # self.diag_cov_fun = diag_cov_fun

    def cholesky(self, x, observed=True, nugget=None, return_eval_also=False):
        """

        U = C.cholesky(x[, observed=True, nugget=None])

        Computes Cholesky factorization of self(x,x).

        :Arguments:

            -   `x`: The input array on which to evaluate the covariance.

            -   `observed`: If 'True', any observations are taken into account
                when computing the Cholesky factor. If not, the unobserved
                version of self is used.

            -   `nugget`: The 'nugget' parameter, which will essentially be
                added to the diagonal of C(x,x) before Cholesky factorizing.
        """

        # Number of points in x.
        N_new = x.shape[0]

        U = self.__call__(x, x, regularize=False, observed=observed)
        if return_eval_also:
            C_eval = U.copy('F')

        if nugget is not None:
            for i in xrange(N_new):
                U[i, i] += nugget[i]

        # print self.params, x.shape, observed, nugget

        info = dpotrf_wrap(U)
        if info > 0:
            raise LinAlgError(
                "Matrix does not appear to be positive definite by row %i. Consider another Covariance subclass, such as NearlyFullRankCovariance." %
                info)

        if return_eval_also:
            return U, C_eval
        else:
            return U

    def continue_cholesky(
            self, x, x_old, U_old, observed=True, nugget=None, return_eval_also=False):
        """

        U = C.continue_cholesky(x, x_old, U_old[, observed=True, nugget=None])

        Computes Cholesky factorization of self(z,z). Assumes the Cholesky
        factorization of self(x_old, x_old) has already been computed.

        :Arguments:

            -   `x`: The input array on which to evaluate the Cholesky factorization.

            -   `x_old`: The input array on which the Cholesky factorization has been
                computed.

            -   `U_old`: The Cholesky factorization of C(x_old, x_old).

            -   `observed`: If 'True', any observations are taken into account
                when computing the Cholesky factor. If not, the unobserved
                version of self is used.

            -   `nugget`: The 'nugget' parameter, which will essentially be
                added to the diagonal of C(x,x) before Cholesky factorizing.
        """

        # Concatenation of the old points and new points.
        xtot = vstack((x_old, x))

        # Number of old points.
        N_old = x_old.shape[0]

        # Number of new points.
        N_new = x.shape[0]

        U_new = self.__call__(x, x, regularize=False, observed=observed)

        # not really implemented yet.
        if nugget is not None:
            for i in xrange(N_new):
                U_new[i, i] += nugget[i]

        U = asmatrix(
            zeros((N_new + N_old,
                   N_old + N_new),
                  dtype=float,
                  order='F'))
        U[:N_old, :N_old] = U_old

        offdiag = self.__call__(
            x=x_old,
            y=x,
            observed=observed,
            regularize=False)
        trisolve(U_old, offdiag, uplo='U', transa='T', inplace=True)
        U[:N_old, N_old:] = offdiag

        U_new -= offdiag.T * offdiag
        if return_eval_also:
            C_eval = U_new.copy('F')

        info = dpotrf_wrap(U_new)
        if info > 0:
            raise LinAlgError(
                "Matrix does not appear to be positive definite by row %i. Consider another Covariance subclass, such as NearlyFullRankCovariance." %
                info)

        U[N_old:, N_old:] = U_new
        if return_eval_also:
            return U, U_new, C_eval
        else:
            return U

    def __call__(self, x, y=None, observed=True,
                 regularize=True, return_Uo_Cxo=False):
        out = Covariance.__call__(
            self,
            x,
            y,
            observed,
            regularize,
            return_Uo_Cxo=return_Uo_Cxo)

        if self.nugget is None:
            return out

        if return_Uo_Cxo:
            out, Uo_Cxo = out
        if x is y:
            for i in xrange(out.shape[0]):
                out[i, i] += self.nugget
        elif y is None:
            out += self.nugget
        if return_Uo_Cxo:
            return out, Uo_Cxo
        else:
            return out

    def observe(self, obs_mesh, obs_V, output_type='r'):
        """
        Observes self on obs_mesh with observation variance obs_V.
        Output_type controls the information returned:

        'r' : returns information needed by Realization objects.
        'o' : returns information needed by function observe.
        's' : returns information needed by the Gaussian process
              submodel.
        """

        # Number of spatial dimensions.
        ndim = obs_mesh.shape[1]

        if self.ndim is not None:
            if not ndim == self.ndim:
                raise ValueError(
                    "Dimension of observation mesh is not equal to dimension of base mesh.")
        else:
            self.ndim = ndim

        # print ndim

        # =====================================
        # = If self hasn't been observed yet: =
        # =====================================
        if not self.observed:

            # If self has not been observed, get the Cholesky factor of self(obs_mesh, obs_mesh)
            # and the side information and store it.

            # Number of observation points so far is 0.
            N_old = 0
            N_new = obs_mesh.shape[0]

            if output_type == 's':
                U, C_eval = self.cholesky(
                    obs_mesh, nugget=obs_V, observed=False, return_eval_also=True)
                U_new = U
            else:
                U = self.cholesky(obs_mesh, nugget=obs_V, observed=False)

            # Upper-triangular Cholesky factor of self(obs_mesh, obs_mesh)
            self.full_Uo = U
            self.Uo = U

            # Pivots.
            piv_new = arange(N_new)
            self.full_piv = piv_new
            self.obs_piv = piv_new

            # Remember full observation mesh.
            self.full_obs_mesh = obs_mesh

            # relevant slice is the positive-definite indices, which get into
            # obs_mesh_*. See documentation.
            relevant_slice = self.obs_piv

            self.obs_mesh = obs_mesh
            self.obs_V = obs_V
            self.obs_len = N_new

        # =======================================
        # = If self has been observed already:  =
        # =======================================
        else:

            # If self has been observed, get the Cholesky factor of the _full_ observation mesh (new
            # and old observations) using continue_cholesky, along with side
            # information, and store it.

            # Number of observations so far.
            N_old = self.full_obs_mesh.shape[0]

            # Number of new observations.
            N_new = obs_mesh.shape[0]

            # Call to self.continue_cholesky.
            if output_type == 's':
                U, U_new, C_eval = self.continue_cholesky(x=obs_mesh,
                                                          x_old=self.full_obs_mesh,
                                                          U_old=self.full_Uo,
                                                          observed=False,
                                                          nugget=obs_V,
                                                          return_eval_also=True)

            else:
                U = self.continue_cholesky(x=obs_mesh,
                                           x_old=self.full_obs_mesh,
                                           U_old=self.full_Uo,
                                           observed=False,
                                           nugget=obs_V)

            # Full Cholesky factor of self(obs_mesh, obs_mesh), where obs_mesh
            # is the combined observation mesh.
            self.full_Uo = U

            # Square upper-triangular Cholesky factor of self(obs_mesh_*,
            # obs_mesh_*). See documentation.
            self.Uo = self.full_Uo

            # Pivots.
            piv_new = arange(N_old + N_new)
            self.obs_piv = piv_new
            self.full_piv = piv_new

            # Concatenate old and new observation meshes.
            self.full_obs_mesh = vstack((self.full_obs_mesh, obs_mesh))
            self.obs_mesh = self.full_obs_mesh
            self.obs_V = hstack((self.obs_V, obs_V))

            # Length of obs_mesh_*.
            self.obs_len = N_old + N_new

        self.observed = True
        # Output expected by Realization
        if output_type == 'r':
            return slice(None, None, None), obs_mesh, self.full_Uo[
                N_old:N_new + N_old, N_old:N_new + N_old], self.full_Uo[:N_old, N_old:N_new + N_old]

        # Ouptut expected by observe
        if output_type == 'o':
            return slice(None, None, None), obs_mesh

        # Output expected by the GP submodel
        if output_type == 's':
            return U_new, C_eval, self.full_Uo[:N_old, N_old:N_new + N_old]



__docformat__ = 'reStructuredText'

__author__ = 'Anand Patil, anand.prabhakar.patil@gmail.com'
__all__ = ['NormApproxMu', 'NormApproxC', 'MAP', 'NormApprox']

from .Node import ZeroProbability
from .Model import Model, Sampler
from numpy import zeros, inner, asmatrix, ndarray
from numpy import reshape, shape, arange, ravel, log, Inf
from numpy.random import normal
from .utils import msqrt, check_type, round_array, logp_of_set
from copy import copy

from . import six
from .six import print_
xrange = six.moves.xrange

try:
    from scipy.optimize import fmin_ncg, fmin, fmin_powell, fmin_cg, fmin_bfgs, fmin_ncg, fmin_l_bfgs_b
    from scipy.misc import derivative
    scipy_imported = True
except ImportError:
    scipy_imported = False


class NormApproxMu(object):

    """
    Returns the mean vector of some variables.

    Usage: If p1 and p2 are array-valued stochastic variables and N is a
    NormApprox or MAP object,

    N.mu(p1,p2)

    will give the approximate posterior mean of the ravelled, concatenated
    values of p1 and p2.
    """

    def __init__(self, owner):
        self.owner = owner

    def __getitem__(self, *stochastics):

        if not self.owner.fitted:
            raise ValueError(
                'NormApprox object must be fitted before mu can be accessed.')

        tot_len = 0

        try:
            for p in stochastics[0]:
                pass
            stochastic_tuple = stochastics[0]
        except:
            stochastic_tuple = stochastics

        for p in stochastic_tuple:
            tot_len += self.owner.stochastic_len[p]

        mu = zeros(tot_len, dtype=float)

        start_index = 0
        for p in stochastic_tuple:
            mu[start_index:(start_index + self.owner.stochastic_len[
                            p])] = self.owner._mu[self.owner._slices[p]]
            start_index += self.owner.stochastic_len[p]

        return mu


class NormApproxC(object):

    """
    Returns the covariance matrix of some variables.

    Usage: If p1 and p2 are array-valued stochastic variables and N is a
    NormApprox or MAP object,

    N.C(p1,p2)

    will give the approximate covariance matrix of the ravelled, concatenated
    values of p1 and p2
    """

    def __init__(self, owner):
        self.owner = owner

    def __getitem__(self, *stochastics):

        if not self.owner.fitted:
            raise ValueError(
                'NormApprox object must be fitted before C can be accessed.')

        tot_len = 0

        try:
            for p in stochastics[0]:
                pass
            stochastic_tuple = stochastics[0]
        except:
            stochastic_tuple = stochastics

        for p in stochastic_tuple:
            tot_len += self.owner.stochastic_len[p]

        C = asmatrix(zeros((tot_len, tot_len)), dtype=float)

        start_index1 = 0
        for p1 in stochastic_tuple:
            start_index2 = 0
            for p2 in stochastic_tuple:
                C[start_index1:(start_index1 + self.owner.stochastic_len[p1]),
                  start_index2:(start_index2 + self.owner.stochastic_len[p2])] = \
                    self.owner._C[
                        self.owner._slices[
                            p1],
                        self.owner._slices[
                            p2]]

                start_index2 += self.owner.stochastic_len[p2]

            start_index1 += self.owner.stochastic_len[p1]

        return C


class MAP(Model):

    """
    N = MAP(input, eps=.001, diff_order = 5)

    Sets all parameters to maximum a posteriori values.

    Useful methods:
    revert_to_max:  Sets all stochastic variables to mean value under normal approximation
    fit:            Finds the normal approximation.

    Useful attributes (after fit() is called):
    logp:               Returns the log-probability of the model
    logp_at_max:        Returns the maximum log-probability of the model
    lnL:                The maximum log likelihood of the model
    len:                The number of free stochastic variables in the model ('k' in AIC and BIC)
    data_len:           The number of datapoints used ('n' in BIC)
    AIC:                Akaike's Information Criterion for the model
    AICc:               Akaike's Information Criterion with correction for small samples
    BIC:                Bayesian Information Criterion for the model

    :Arguments:
    input: As for Model
    eps: 'h' for computing numerical derivatives. May be a dictionary keyed by stochastic variable
      as well as a scalar.
    diff_order: The order of the approximation used to compute derivatives.

    :SeeAlso: Model, EM, Sampler, scipy.optimize
    """

    def __init__(self, input=()

, eps=.001, diff_order=5, verbose=-1):
        if not scipy_imported:
            raise ImportError(
                'Scipy must be installed to use NormApprox and MAP.')

        Model.__init__(self, input, verbose=verbose)

        # Allocate memory for internal traces and get stochastic slices
        self._slices = {}
        self.len = 0
        self.stochastic_len = {}
        self.fitted = False

        self.stochastic_list = list(self.stochastics)
        self.N_stochastics = len(self.stochastic_list)
        self.stochastic_indices = []
        self.stochastic_types = []
        self.stochastic_type_dict = {}

        for i in xrange(len(self.stochastic_list)):

            stochastic = self.stochastic_list[i]

            # Check types of all stochastics.
            type_now = check_type(stochastic)[0]
            self.stochastic_type_dict[stochastic] = type_now

            if not type_now is float:
                print_("Warning: Stochastic " + stochastic.__name__ + "'s value is neither numerical nor array with " +
                       "floating-point dtype. Recommend fitting method fmin (default).")

            # Inspect shapes of all stochastics and create stochastic slices.
            if isinstance(stochastic.value, ndarray):
                self.stochastic_len[stochastic] = len(ravel(stochastic.value))
            else:
                self.stochastic_len[stochastic] = 1
            self._slices[
                stochastic] = slice(
                    self.len,
                    self.len +
                    self.stochastic_len[
                        stochastic])
            self.len += self.stochastic_len[stochastic]

            # Record indices that correspond to each stochastic.
            for j in range(len(ravel(stochastic.value))):
                self.stochastic_indices.append((stochastic, j))
                self.stochastic_types.append(type_now)

        self.data_len = 0
        for datum in self.observed_stochastics:
            self.data_len += len(ravel(datum.value))

        # Unpack step
        self.eps = zeros(self.len, dtype=float)
        if isinstance(eps, dict):
            for stochastic in self.stochastics:
                self.eps[self._slices[stochastic]] = eps[stochastic]
        else:
            self.eps[:] = eps

        self.diff_order = diff_order

        self._len_range = arange(self.len)

        # Initialize gradient and Hessian matrix.
        self.grad = zeros(self.len, dtype=float)
        self.hess = asmatrix(zeros((self.len, self.len), dtype=float))

        self._mu = None

        # Initialize NormApproxMu object.
        self.mu = NormApproxMu(self)

        def func_for_diff(val, index):
            """
            The function that gets passed to the derivatives.
            """
            self[index] = val
            return self.i_logp(index)

        self.func_for_diff = func_for_diff

    def fit(self, method='fmin_powell', iterlim=1000, tol=.0001, verbose=0,
            no_callback=False, **kwargs):
        """
        N.fit(method='fmin_powell', iterlim=1000, tol=.001):

        Causes the normal approximation object to fit itself.

        method: May be one of the following, from the scipy.optimize package:
            -fmin_l_bfgs_b
            -fmin_ncg
            -fmin_cg
            -fmin_powell
            -fmin

        no_callback: Boolean indicating whether or not to use a callback
        function. If True and a callback keyword is provided in kwargs, then
        the user-supplied callback will be used. Otherwise, if False,
        and verbose > 0, a default callback will print iteration progress.

        The kwargs are passed to the scipy.optimize functions. See there
        for more information.
        """
        self.tol = tol
        self.method = method
        self.verbose = verbose

        p = zeros(self.len, dtype=float)
        for stochastic in self.stochastics:
            p[self._slices[stochastic]] = ravel(stochastic.value)

        if not self.method == 'newton':
            if not scipy_imported:
                raise ImportError('Scipy is required to use EM and NormApprox')

        default_callback = (verbose > 0 and not no_callback)
        if default_callback and 'callback' in kwargs:
            raise ValueError("For user-provided callback and verbose output"
                             " set use_callback to True")

        if default_callback:
            def callback(p):
                try:
                    print_('Current log-probability : %f' % self.logp)
                except ZeroProbability:
                    print_('Current log-probability : %f' % -Inf)
        elif 'callback' in kwargs:
            callback = kwargs.pop('callback')
        else:
            def callback(p):
                pass

        if self.method == 'fmin_ncg':
            p = fmin_ncg(f=self.func,
                         x0=p,
                         fprime=self.gradfunc,
                         fhess=self.hessfunc,
                         epsilon=self.eps,
                         maxiter=iterlim,
                         callback=callback,
                         avextol=tol,
                         disp=verbose,
                         **kwargs)

        elif self.method == 'fmin':

            p = fmin(func=self.func,
                     x0=p,
                     callback=callback,
                     maxiter=iterlim,
                     ftol=tol,
                     disp=verbose, **kwargs)

        elif self.method == 'fmin_powell':
            p = fmin_powell(func=self.func,
                            x0=p,
                            callback=callback,
                            maxiter=iterlim,
                            ftol=tol,
                            disp=verbose, **kwargs)

        elif self.method == 'fmin_cg':
            p = fmin_cg(f=self.func, x0=p,
                        fprime=self.gradfunc,
                        epsilon=self.eps,
                        callback=callback,
                        maxiter=iterlim,
                        gtol=tol,
                        disp=verbose, **kwargs)

        elif self.method == 'fmin_l_bfgs_b':
            from scipy import __version__ as sp_version
            from distutils.version import LooseVersion
            if LooseVersion(sp_version) >= LooseVersion('0.12.0'):
                p = fmin_l_bfgs_b(func=self.func,
                                  x0=p,
                                  fprime=self.gradfunc,
                                  epsilon=self.eps,
                                  callback=callback,
                                  pgtol=tol,
                                  iprint=verbose - 1,
                                  **kwargs)[0]
            else:
                if verbose > 0:
                    from warnings import warn
                    warn("Callbacks are not available for fmin_l_bfgs_b in "
                         "SciPy < 0.12.0. Optimization progress will not be"
                         "displayed.", UserWarning)
                p = fmin_l_bfgs_b(func=self.func,
                                  x0=p,
                                  fprime=self.gradfunc,
                                  epsilon=self.eps,
                                  pgtol=tol,
                                  iprint=verbose - 1,
                                  **kwargs)[0]
        else:
            raise ValueError('Method unknown.')

        self._set_stochastics(p)
        self._mu = p

        try:
            self.logp_at_max = self.logp
        except:
            raise RuntimeError(
                'Posterior probability optimization converged to value with zero probability.')

        lnL = sum([x.logp for x in self.observed_stochastics]
                  )  # log-likelihood of observed stochastics
        self.lnL = lnL
        try:
            self.AIC = 2. * (self.len - lnL)  # 2k - 2 ln(L)
            self.AICc = self.AIC + ((2 * self.len * (self.len + 1)) / float(self.data_len - self.len - 1))
        except Exception as e:
            print('Cannot calculate AIC:', e)
            self.AICc = self.AIC = -Inf
        try:
            self.BIC = self.len * log(
                self.data_len) - 2. * lnL  # k ln(n) - 2 ln(L)
        except FloatingPointError as e:
            print('Cannot calculate BIC:', e)
            self.BIC = -Inf

        self.fitted = True

    def func(self, p):
        """
        The function that gets passed to the optimizers.
        """
        self._set_stochastics(p)
        try:
            return -1. * self.logp
        except ZeroProbability:
            return Inf

    def gradfunc(self, p):
        """
        The gradient-computing function that gets passed to the optimizers,
        if needed.
        """
        self._set_stochastics(p)
        for i in xrange(self.len):
            self.grad[i] = self.diff(i)

        return -1 * self.grad

    def _set_stochastics(self, p):
        for stochastic in self.stochastics:
            if self.stochastic_type_dict[stochastic] is int:
                stochastic.value = round_array(
                    reshape(ravel(p)[self._slices[stochastic]],
                            shape(stochastic.value)))
            else:
                stochastic.value = reshape(
                    ravel(p)[self._slices[stochastic]],
                    shape(stochastic.value))

    def __setitem__(self, index, value):
        p, i = self.stochastic_indices[index]
        val = ravel(p.value).copy()
        val[i] = value
        p.value = reshape(val, shape(p.value))

    def __getitem__(self, index):
        p, i = self.stochastic_indices[index]
        val = ravel(p.value)
        return val[i]

    def i_logp(self, index):
        """
        Evaluates the log-probability of the Markov blanket of
        a stochastic owning a particular index.
        """
        all_relevant_stochastics = set()
        p, i = self.stochastic_indices[index]
        try:
            return p.logp + logp_of_set(p.extended_children)
        except ZeroProbability:
            return -Inf

    def diff(self, i, order=1):
        """
        N.diff(i, order=1)

        Derivative wrt index i to given order.
        """

        old_val = copy(self[i])
        d = derivative(
            func=self.func_for_diff,
            x0=old_val,
            dx=self.eps[i],
            n=order,
            args=[i],
            order=self.diff_order)
        self[i] = old_val
        return d

    def diff2(self, i, j):
        """
        N.diff2(i,j)

        Mixed second derivative. Differentiates wrt both indices.
        """

        old_val = copy(self[j])

        if not self.stochastic_indices[i][
                0] in self.stochastic_indices[j][0].moral_neighbors:
            return 0.

        def diff_for_diff(val):
            self[j] = val
            return self.diff(i)

        d = derivative(
            func=diff_for_diff,
            x0=old_val,
            dx=self.eps[j],
            n=1,
            order=self.diff_order)

        self[j] = old_val
        return d

    def grad_and_hess(self):
        """
        Computes self's gradient and Hessian. Used if the
        optimization method for a NormApprox doesn't
        use gradients and hessians, for instance fmin.
        """
        for i in xrange(self.len):

            di = self.diff(i)
            self.grad[i] = di
            self.hess[i, i] = self.diff(i, 2)

            if i < self.len - 1:

                for j in xrange(i + 1, self.len):
                    dij = self.diff2(i, j)

                    self.hess[i, j] = dij
                    self.hess[j, i] = dij

    def hessfunc(self, p):
        """
        The Hessian function that will be passed to the optimizer,
        if needed.
        """
        self._set_stochastics(p)
        for i in xrange(self.len):

            di = self.diff(i)
            self.hess[i, i] = self.diff(i, 2)

            if i < self.len - 1:

                for j in xrange(i + 1, self.len):
                    dij = self.diff2(i, j)

                    self.hess[i, j] = dij
                    self.hess[j, i] = dij

        return -1. * self.hess

    def revert_to_max(self):
        """
        N.revert_to_max()

        Sets all N's stochastics to their MAP values.
        """
        try:
            self._set_stochastics([self.mu[s] for s in self.stochastics])
        except KeyError:
            self._set_stochastics(self.mu[self.stochastics])


class NormApprox(MAP, Sampler):

    """
    N = NormApprox(input, db='ram', eps=.001, diff_order = 5, **kwds)

    Normal approximation to the posterior of a model.

    Useful methods:
    draw:           Draws values for all stochastic variables using normal approximation
    revert_to_max:  Sets all stochastic variables to mean value under normal approximation
    fit:            Finds the normal approximation.

    Useful attributes (after fit() is called):
    mu[p1, p2, ...]:    Returns the posterior mean vector of stochastic variables p1, p2, ...
    C[p1, p2, ...]:     Returns the posterior covariance of stochastic variables p1, p2, ...
    logp:               Returns the log-probability of the model
    logp_at_max:        Returns the maximum log-probability of the model
    len:                The number of free stochastic variables in the model ('k' in AIC and BIC)
    data_len:           The number of datapoints used ('n' in BIC)
    AIC:                Akaike's Information Criterion for the model
    BIC:                Bayesian Information Criterion for the model

    :Arguments:
    input: As for Model
    db: A database backend
    eps: 'h' for computing numerical derivatives. May be a dictionary keyed by stochastic variable
      as well as a scalar.
    diff_order: The order of the approximation used to compute derivatives.

    :SeeAlso: Model, EM, Sampler, scipy.optimize
    """

    def __init__(self, input=()

, db='ram', eps=.001, diff_order=5, **kwds):
        if not scipy_imported:
            raise ImportError(
                'Scipy must be installed to use NormApprox and MAP.')

        MAP.__init__(self, input, eps, diff_order)

        Sampler.__init__(self, input, db, reinit_model=False, **kwds)
        self.C = NormApproxC(self)

    def fit(self, *args, **kwargs):
        MAP.fit(self, *args, **kwargs)
        self.fitted = False
        self.grad_and_hess()

        self._C = -1. * self.hess.I
        self._sig = msqrt(self._C).T

        self.fitted = True

    def draw(self):
        """
        N.draw()

        Sets all N's stochastics to random values drawn from
        the normal approximation to the posterior.
        """
        devs = normal(size=self._sig.shape[1])
        p = inner(self._sig, devs) + self._mu
        self._set_stochastics(p)



from numpy.testing import *
from pymc.gp import *
from pymc.gp.cov_funs import matern
from numpy import *
from copy import copy
from .test_mean import x, y

C = Covariance(
    eval_fun=matern.euclidean,
    diff_degree=1.4,
    amp=.4,
    scale=1.)
D = copy(C)


class test_cov(TestCase):

    def test(self):

        assert(C(x).shape == x.shape)
        assert(C(x, x).shape == (len(x), len(x)))

        assert(D(y).shape == (y.shape[0],))
        assert(D(y, y).shape == (y.shape[0], y.shape[0]))



"""
pymc.NumpyDeterministics
"""

__docformat__ = 'reStructuredText'
from . import PyMCObjects as pm
import numpy as np
from numpy import sum, ones, zeros, ravel, shape, size, newaxis
from .utils import find_element, get_signature
import inspect

from . import six
xrange = six.moves.xrange

# accumulations
_boolean_accumulation_deterministics = ['any', 'all']
_accumulation_deterministics = ['sum']  # ['sum', 'prod']


# transformations (broadcasted)
_generic = ['abs', 'exp', 'log', 'sqrt', 'expm1', 'log1p']
_trig = ['sin', 'cos', 'tan', 'arcsin', 'arccos', 'arctan']
_hyp_trig = ['sinh', 'cosh', 'tanh', 'arcsinh', 'arccosh', 'arctanh']
_transformation_deterministics = _generic + _trig + _hyp_trig
_misc_funcs1 = ['arctan2', 'hypot']

__all__ = _accumulation_deterministics  + _boolean_accumulation_deterministics + \
    _transformation_deterministics + _misc_funcs1


def deterministic_from_funcs(
        name, eval, jacobians={}, jacobian_formats={}, dtype=np.float, mv=False):
    """
    Return a Stochastic subclass made from a particular distribution.

    :Parameters:
      name : string
        The name of the new class.
      jacobians : function
        The log-probability function.
      random : function
        The random function
      dtype : numpy dtype
        The dtype of values of instances.
      mv : boolean
        A flag indicating whether this class represents
        array-valued variables.

    """

    (args, defaults) = get_signature(eval)
    parent_names = args[0:]
    try:
        parents_default = dict(zip(args[-len(defaults):], defaults))
    except TypeError:  # No parents at all.
        parents_default = {}

    # Build docstring from distribution
    docstr = name[0] + ' = ' + name + '('.join(parent_names) + ')\n\n'
    docstr += 'Deterministic variable with ' + name + \
        ' distribution.\nParents are: ' + ', '.join(parent_names) + '.\n\n'
    docstr += 'Docstring of evaluatio function:\n'
    docstr += eval.__doc__

    return new_deterministic_class(
        dtype, name, parent_names, parents_default, docstr, eval, jacobians, jacobian_formats)


def new_deterministic_class(*new_class_args):
    """
    Returns a new class from a distribution.

    :Parameters:
      dtype : numpy dtype
        The dtype values of instances of this class.
      name : string
        Name of the new class.
      parent_names : list of strings
        The labels of the parents of this class.
      parents_default : list
        The default values of parents.
      docstr : string
        The docstring of this class.
      eval : function
        The function for this class.
      jacobians : dictionary of functions
        The dictionary of jacobian functions for the class
      jacobian_formats : dictionary of strings
        A dictionary indicating the format of each jacobian function
    """

    (dtype, name, parent_names, parents_default, docstr,
     eval, jacobians, jacobian_formats) = new_class_args

    class new_class(pm.Deterministic):
        __doc__ = docstr

        def __init__(self, *args, **kwds):
            (dtype, name, parent_names, parents_default, docstr,
             eval, jacobians, jacobian_formats) = new_class_args
            parents = parents_default

            # Figure out what argument names are needed.
            arg_keys = ['parents', 'trace', 'doc', 'debug', 'plot', 'verbose']
            arg_vals = [parents, False, True, None, False, -1]

            arg_dict_out = dict(zip(arg_keys, arg_vals))
            args_needed = parent_names + arg_keys[2:]

            # Sort positional arguments
            for i in xrange(len(args)):
                try:
                    k = args_needed.pop(0)
                    if k in parent_names:
                        parents[k] = args[i]
                    else:
                        arg_dict_out[k] = args[i]
                except:
                    raise ValueError(
                        'Too many positional arguments provided. Arguments for class ' + self.__class__.__name__ + ' are: ' + str(
                            all_args_needed))

            # Sort keyword arguments
            for k in args_needed:
                if k in parent_names:
                    try:
                        parents[k] = kwds.pop(k)
                    except:
                        if k in parents_default:
                            parents[k] = parents_default[k]
                        else:
                            raise ValueError('No value given for parent ' + k)
                elif k in arg_dict_out.keys():
                    try:
                        arg_dict_out[k] = kwds.pop(k)
                    except:
                        pass

            # Remaining unrecognized arguments raise an error.
            if len(kwds) > 0:
                raise TypeError(
                    'Keywords ' + str(
                        kwds.keys(
                        )) + ' not recognized. Arguments recognized are ' + str(
                        args_needed))

            # Call base class initialization method
            if arg_dict_out.pop('debug'):
                pass
            else:
                parent_strs = []
                for key in parents.keys():
                    parent_strs.append(str(key))

                instance_name = name + '(' + ','.join(parent_strs) + ')'

                pm.Deterministic.__init__(
                    self,
                    name=instance_name,
                    eval=eval,
                    jacobians=jacobians,
                    jacobian_formats=jacobian_formats,
                    dtype=dtype,
                    **arg_dict_out)

    new_class.__name__ = name
    new_class.parent_names = parent_names

    return new_class


_sum_hist = {}


def sum_jacobian_a(a, axis):
    try:
        return _sum_hist[shape(a)]
    except KeyError:
        j = ones(shape(a))
        _sum_hist[shape(a)] = j
        return j

sum_jacobians = {'a': sum_jacobian_a}

abs_jacobians = {'x': lambda x: np.sign(x)}
exp_jacobians = {'x': lambda x: np.exp(x)}
log_jacobians = {'x': lambda x: 1.0 / x}
sqrt_jacobians = {'x': lambda x: .5 * x ** -.5}
hypot_jacobians = {'x1': lambda x1, x2: (x1 ** 2 + x2 ** 2) ** -.5 * x1,
                   'x2': lambda x1, x2: (x1 ** 2 + x2 ** 2) ** -.5 * x2}
expm1_jacobians = exp_jacobians
log1p_jacobians = {'x': lambda x: 1.0 / (1.0 + x)}

sin_jacobians = {'x': lambda x: np.cos(x)}
cos_jacobians = {'x': lambda x: -np.sin(x)}
tan_jacobians = {'x': lambda x: 1 + np.tan(x) ** 2}

arcsin_jacobians = {'x': lambda x: (1.0 - x ** 2) ** -.5}
arccos_jacobians = {'x': lambda x: -(1.0 - x ** 2) ** -.5}
arctan_jacobians = {'x': lambda x: 1.0 / (1.0 + x ** 2)}
arctan2_jacobians = {'x1': lambda x1, x2: x2 / (x2 ** 2 + x1 ** 2),
                     'x2': lambda x1, x2: -x1 / (x2 ** 2 + x1 ** 2)}
# found in www.math.smith.edu/phyllo/Assets/pdf/findcenter.pdf p21

sinh_jacobians = {'x': lambda x: np.cosh(x)}
cosh_jacobians = {'x': lambda x: np.sinh(x)}
tanh_jacobians = {'x': lambda x: 1.0 - np.tanh(x) ** 2}

arcsinh_jacobians = {'x': lambda x: (1 + x ** 2) ** -.5}
arccosh_jacobians = {'x': lambda x: (x + 1) ** -.5 * (x - 1.0) ** -.5}
arctanh_jacobians = {'x': lambda x: 1.0 / (1 - x ** 2)}


def wrap_function_accum(function):
    def wrapped_function(a, axis=None):
        return function(a, axis)
    wrapped_function.__doc__ = function.__doc__

    return wrapped_function

for function_name in _accumulation_deterministics:
    wrapped_function = wrap_function_accum(
        find_element(function_name,
                     np,
                     error_on_fail=True))

    jacobians = find_element(
        function_name +
        "_jacobians",
        locals(
        ),
        error_on_fail=True)

    locals(
    )[function_name] = deterministic_from_funcs(function_name,
                                                wrapped_function,
                                                jacobians,
                                                jacobian_formats={'a': 'accumulation_operation'})


for function_name in _boolean_accumulation_deterministics:
    wrapped_function = wrap_function_accum(
        find_element(function_name,
                     np,
                     error_on_fail=True))

    locals()[function_name] = deterministic_from_funcs(
        function_name, wrapped_function)


def wrapped_function_trans(function):
    def wrapped_function(x):
        return function(x)
    wrapped_function.__doc__ = function.__doc__

    return wrapped_function

for function_name in _transformation_deterministics:
    wrapped_function = wrapped_function_trans(
        find_element(function_name,
                     np,
                     error_on_fail=True))

    jacobians = find_element(
        function_name +
        "_jacobians",
        locals(
        ),
        error_on_fail=True)
    locals(
    )[function_name] = deterministic_from_funcs(function_name,
                                                wrapped_function,
                                                jacobians,
                                                jacobian_formats={'x': 'transformation_operation'})


def wrap_function_misc1(function):
    def wrapped_function(x1, x2):
        return function(x1, x2)
    wrapped_function.__doc__ = function.__doc__

    return wrapped_function

for function_name in _misc_funcs1:
    wrapped_function = wrap_function_misc1(
        find_element(function_name,
                     np,
                     error_on_fail=True))

    jacobians = find_element(
        function_name +
        "_jacobians",
        locals(
        ),
        error_on_fail=True)

    locals(
    )[function_name] = deterministic_from_funcs(
        function_name, wrapped_function, jacobians, jacobian_formats={
            'x1': 'broadcast_operation',
            'x2': 'broadcast_operation'})



# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'

__all__ = ['NearlyFullRankCovariance']

from numpy import *
from numpy.linalg import cholesky, LinAlgError
from .GPutils import regularize_array, trisolve
from .linalg_utils import diag_call
from .incomplete_chol import ichol_full
from .Covariance import Covariance

from pymc import six
xrange = six.moves.xrange


class NearlyFullRankCovariance(Covariance):

    """
    C=NearlyFullRankCovariance(eval_fun, relative_precision, **params)

    A GP covariance. Good for situations where self's evaluation
    on observation locations is nearly full-rank, but not quite. Evaluation
    of the matrix can be parallelized, but Cholesky decompositions are done
    using a serial incomplete algorithm.

    :Arguments:

        -   `eval_fun`: A function that takes either a single value x or two values x and y,
            followed by an arbitrary number of keyword parameters. x and y will be of shape
            (n,n_dim), where n is any integer and n_dim is the dimensionality of the space, or
            shape (n). In the latter case n_dim should be assumed to be 1.

        -   `params`: Parameters to be passed to eval_fun.

        -   `relative_precision`: See documentation.


    :SeeAlso: Mean, BasisCovariance, SeparableBasisCovariance, Realization, observe
    """

    def __init__(self, eval_fun, relative_precision=1.0E-15, **params):
        Covariance.__init__(
            self,
            eval_fun,
            relative_precision,
            rank_limit=0,
            **params)

    def cholesky(self, x, apply_pivot=True, observed=True,
                 nugget=None, regularize=True, rank_limit=0):
        """

        U = C.cholesky(x[, observed=True, nugget=None, rank_limit=0])

        {'pivots': piv, 'U': U} = \
        C.cholesky(x, apply_pivot = False[, observed=True, nugget=None])

        Computes incomplete Cholesky factorization of self(x,x).


        :Arguments:

            -   `x`: The input array on which to evaluate the covariance.

            -   `apply_pivot`: A flag. If it's set to 'True', it returns a
                matrix U (not necessarily triangular) such that U.T*U=C(x,x).
                If it's set to 'False', the return value is a dictionary.
                Item 'pivots' is a vector of pivots, and item 'U' is an
                upper-triangular matrix (not necessarily square) such that
                U[:,argsort(piv)].T * U[:,argsort(piv)] = C(x,x).

            -   `observed`: If 'True', any observations are taken into account
                when computing the Cholesky factor. If not, the unobserved
                version of self is used.

            -   `nugget`: The 'nugget' parameter, which will essentially be
                added to the diagonal of C(x,x) before Cholesky factorizing.

            -   `rank_limit`: If rank_limit > 0, the factor will have at most
                rank_limit rows.
        """
        if rank_limit > 0:
            raise ValueError(
                'NearlyFullRankCovariance does not accept a rank_limit argument. Use Covariance instead.')

        if regularize:
            x = regularize_array(x)

        # Number of points in x.
        N_new = x.shape[0]

        # Special fast version for single points.
        if N_new == 1:
            V = self.__call__(x, regularize=False, observed=observed)
            if nugget is not None:
                V += nugget
            U = asmatrix(sqrt(V))
            # print U
            if not apply_pivot:
                return {'pivots': array([0]), 'U': U}
            else:
                return U

        C = self.__call__(x, x, regularize=False, observed=observed)
        if nugget is not None:
            for i in xrange(N_new):
                C[i, i] += nugget[i]

        # =======================================
        # = Call to Fortran function ichol_full =
        # =======================================
        U, m, piv = ichol_full(c=C, reltol=self.relative_precision)

        U = asmatrix(U)

        # Arrange output matrix and return.
        if m < 0:
            raise ValueError(
                "Matrix does not appear to be positive semidefinite")

        if not apply_pivot:
            # Useful for self.observe and Realization.__call__. U is upper
            # triangular.
            U = U[:m, :]
            return {'pivots': piv, 'U': U}

        else:
            # Useful for users. U.T*U = C(x,x)
            return U[:m, argsort(piv)]

    def continue_cholesky(self, x, x_old, chol_dict_old, apply_pivot=True,
                          observed=True, nugget=None, regularize=True, assume_full_rank=False, rank_limit=0):
        """

        U = C.continue_cholesky(x, x_old, chol_dict_old[, observed=True, nugget=None,
                rank_limit=0])


        Returns {'pivots': piv, 'U': U}


        Computes incomplete Cholesky factorization of self(z,z). Here z is the
        concatenation of x and x_old. Assumes the Cholesky factorization of
        self(x_old, x_old) has already been computed.


        :Arguments:

            -   `x`: The input array on which to evaluate the Cholesky factorization.

            -   `x_old`: The input array on which the Cholesky factorization has been
                computed.

            -   `chol_dict_old`: A dictionary with kbasis_ys ['pivots', 'U']. Would be the
                output of either this method or C.cholesky().

            -   `apply_pivot`: A flag. If it's set to 'True', it returns a
                matrix U (not necessarily triangular) such that U.T*U=C(x,x).
                If it's set to 'False', the return value is a dictionary.
                Item 'pivots' is a vector of pivots, and item 'U' is an
                upper-triangular matrix (not necessarily square) such that
                U[:,argsort(piv)].T * U[:,argsort(piv)] = C(x,x).

            -   `observed`: If 'True', any observations are taken into account
                when computing the Cholesky factor. If not, the unobserved
                version of self is used.

            -   `nugget`: The 'nugget' parameter, which will essentially be
                added to the diagonal of C(x,x) before Cholesky factorizing.

            -   `rank_limit`: If rank_limit > 0, the factor will have at most
                rank_limit rows.
        """
        if regularize:
            x = regularize_array(x)

        if rank_limit > 0:
            raise ValueError(
                'NearlyFullRankCovariance does not accept a rank_limit argument. Use Covariance instead.')

        if rank_limit > 0:
            raise ValueError(
                'NearlyFullRankCovariance does not accept a rank_limit argument. Use Covariance instead.')

        # Concatenation of the old points and new points.
        xtot = vstack((x_old, x))

        # Extract information from chol_dict_old.
        U_old = chol_dict_old['U']
        m_old = U_old.shape[0]
        piv_old = chol_dict_old['pivots']

        # Number of old points.
        N_old = x_old.shape[0]

        # Number of new points.
        N_new = x.shape[0]

        # Compute off-diagonal part of Cholesky factor
        offdiag = self.__call__(
            x=x_old[
                piv_old[
                    :m_old],
                :],
            y=x,
            observed=observed,
            regularize=False)
        trisolve(U_old[:, :m_old], offdiag, uplo='U', transa='T', inplace=True)

        # Compute new diagonal part of Cholesky factor
        C_new = self.__call__(x=x, y=x, observed=observed, regularize=False)
        if nugget is not None:
            for i in xrange(N_new):
                C_new[i, i] += nugget[i]
        C_new -= offdiag.T * offdiag
        if not assume_full_rank:
            U_new, m_new, piv_new = ichol_full(
                c=C_new, reltol=self.relative_precision)
        else:
            U_new = cholesky(C_new).T
            m_new = U_new.shape[0]
            piv_new = arange(m_new)
        U_new = asmatrix(U_new[:m_new, :])
        U = asmatrix(
            zeros(
                (m_new +
                 m_old,
                 N_old +
                 N_new),
                dtype=float,
                order='F'))

        # Top portion of U
        U[:m_old, :m_old] = U_old[:, :m_old]
        U[:m_old, N_new + m_old:] = U_old[:, m_old:]
        offdiag = offdiag[:, piv_new]
        U[:m_old, m_old:N_new + m_old] = offdiag

        # Lower portion of U
        U[m_old:, m_old:m_old + N_new] = U_new
        if m_old < N_old and m_new > 0:
            offdiag_lower = self.__call__(x=x[piv_new[:m_new], :],
                                          y=x_old[piv_old[m_old:], :], observed=observed, regularize=False)
            offdiag_lower -= offdiag[:, :m_new].T * U[:m_old, m_old + N_new:]
            trisolve(
                U_new[
                    :,
                    :m_new],
                offdiag_lower,
                uplo='U',
                transa='T',
                inplace=True)
            U[m_old:, m_old + N_new:] = offdiag_lower

        # Rank and pivots
        m = m_old + m_new
        piv = hstack((piv_old[:m_old], piv_new + N_old, piv_old[m_old:]))

        # Arrange output matrix and return.
        if m < 0:
            raise ValueError('Matrix does not appear positive semidefinite.')

        if not apply_pivot:
            # Useful for self.observe. U is upper triangular.
            if assume_full_rank:
                return {'pivots': piv, 'U': U, 'C_eval': C_new, 'U_new': U_new}
            else:
                return {'pivots': piv, 'U': U}

        else:
            # Useful for the user. U.T * U = C(x,x).
            return U[:, argsort(piv)]


# def clean_array(A):
#     for i in xrange(A.shape[0]):
#         for j in xrange(A.shape[1]):
#             if abs(A[i,j])<1e-10:
#                 A[i,j]=0
#
# if __name__=='__main__':
#     from cov_funs import matern
#     C = NearlyFullRankCovariance(eval_fun = matern.euclidean, diff_degree = 1.4, amp = .4, scale = 1.)
#     first=C.cholesky([1,1,2],apply_pivot=False)
#     second=C.continue_cholesky([3,4],regularize_array([1,1,2]),first)
#     clean_array(first['U'])
#     clean_array(second)
#     print second.T*second
#     print
#     print C([1,1,2,3,4],[1,1,2,3,4])



from __future__ import with_statement
"""
TXT database module

Store the traces in ASCII files.

For each chain, a directory named `Chain_#` is created. In this directory,
one file per tallyable object is created containing the values of the object.

Implementation Notes
--------------------
The NumPy arrays are saved and loaded using NumPy's `loadtxt` and `savetxt`
functions.

Changeset
---------
Nov. 30, 2007: Implemented load function. DH
Oct. 24, 2008: Implemented savestate. Implemented parallel chain tallying. DH
Oct. 1, 2009: Added support for multidimensional arrays.
"""


from . import base, ram
import os
import datetime
import shutil
import re
import numpy as np
from numpy import array
import string

from pymc import six
from pymc.six import print_


__all__ = ['Trace', 'Database', 'load']

CHAIN_NAME = 'Chain_%d'


class Trace(ram.Trace):

    """Txt Trace Class.

    Store the trace in a ASCII file located in one directory per chain.

    dbname/
      Chain_0/
        <object name>.txt
        <object name>.txt
        ...
      Chain_1/
        <object name>.txt
        <object name>.txt
        ...
      ...
    """

    def _finalize(self, chain):
        """Write the trace to an ASCII file.

        :Parameter:
        chain : int
          The chain index.
        """
        path = os.path.join(
            self.db._directory,
            self.db.get_chains()[chain],
            self.name + '.txt')
        arr = self.gettrace(chain=chain)

        # Following numpy's example.
        if six.PY3:
            mode = 'wb'
        else:
            mode = 'w'
        with open(path, mode) as f:
            f.write(six.b('# Variable: %s\n' % self.name))
            f.write(six.b('# Sample shape: %s\n' % str(arr.shape)))
            f.write(six.b('# Date: %s\n' % datetime.datetime.now()))
            np.savetxt(f, arr.reshape((-1, arr[0].size)), delimiter=',')


class Database(base.Database):

    """Txt Database class."""

    def __init__(self, dbname=None, dbmode='a'):
        """Create a Txt Database.

        :Parameters:
        dbname : string
          Name of the directory where the traces are stored.
        dbmode : {a, r, w}
          Opening mode: a:append, w:write, r:read.
        """
        self.__name__ = 'txt'
        self._directory = dbname
        self.__Trace__ = Trace
        self.mode = dbmode

        self.trace_names = []
        # A list of sequences of names of the objects to tally.
        self._traces = {}  # A dictionary of the Trace objects.
        self.chains = 0

        if os.path.exists(self._directory):
            if dbmode == 'w':
                shutil.rmtree(self._directory)
                os.mkdir(self._directory)
        else:
            os.mkdir(self._directory)

    def get_chains(self):
        """Return an ordered list of the `Chain_#` directories in the db
        directory."""
        chains = []
        try:
            content = os.listdir(self._directory)
            for c in content:
                if os.path.isdir(os.path.join(self._directory, c)) and c.startswith(
                        CHAIN_NAME[:-2]):
                    chains.append(c)
        except:
            pass
        chains.sort()
        return chains

    def _initialize(self, funs_to_tally, length):
        """Create folder to store simulation results."""

        dir = os.path.join(self._directory, CHAIN_NAME % self.chains)
        os.mkdir(dir)

        base.Database._initialize(self, funs_to_tally, length)

    def savestate(self, state):
        """Save the sampler's state in a state.txt file."""
        oldstate = np.get_printoptions()
        np.set_printoptions(threshold=1e6)
        try:
            with open(os.path.join(self._directory, 'state.txt'), 'w') as f:
                print_(state, file=f)
        finally:
            np.set_printoptions(**oldstate)


def load(dirname):
    """Create a Database instance from the data stored in the directory."""
    if not os.path.exists(dirname):
        raise AttributeError('No txt database named %s' % dirname)

    db = Database(dirname, dbmode='a')
    chain_folders = [os.path.join(dirname, c) for c in db.get_chains()]
    db.chains = len(chain_folders)

    data = {}
    for chain, folder in enumerate(chain_folders):
        files = os.listdir(folder)
        funnames = funname(files)
        db.trace_names.append(funnames)
        for file in files:
            name = funname(file)
            if name not in data:
                data[
                    name] = {
                }  # This could be simplified using "collections.defaultdict(dict)". New in Python 2.5
            # Read the shape information
            with open(os.path.join(folder, file)) as f:
                f.readline()
                shape = eval(f.readline()[16:])
                data[
                    name][
                        chain] = np.loadtxt(
                            os.path.join(
                                folder,
                                file),
                            delimiter=',').reshape(
                                shape)
                f.close()

    # Create the Traces.
    for name, values in six.iteritems(data):
        db._traces[name] = Trace(name=name, value=values, db=db)
        setattr(db, name, db._traces[name])

    # Load the state.
    statefile = os.path.join(dirname, 'state.txt')
    if os.path.exists(statefile):
        with open(statefile, 'r') as f:
            db._state_ = eval(f.read())
    else:
        db._state_ = {}

    return db


def funname(file):
    """Return variable names from file names."""
    if isinstance(file, str):
        files = [file]
    else:
        files = file
    bases = [os.path.basename(f) for f in files]
    names = [os.path.splitext(b)[0] for b in bases]
    if isinstance(file, str):
        return names[0]
    else:
        return names



import pymc as pm
from numpy.testing import *
import numpy as np
import nose
import sys

from pymc import six
xrange = six.moves.xrange

DIR = 'testresults'


def mymodel():
    mu = pm.Normal('mu', 0, 1)
    N = [pm.Normal('N_%i' % i, mu, 1) for i in xrange(3)]
    z1 = pm.Lambda('z1', lambda n=N: np.sum(n))
    z2 = pm.Lambda('z2', lambda n=N: np.sum(n))

    @pm.potential
    def y(z1=z1, z2=z2, mu=mu):
        return 0
    return mu, N, z1, z2, y


def powerset(seq):
    """
    Returns all the subsets of this set. This is a generator.

    From http://blog.technomancy.org/2009/3/17/a-powerset-generator-in-python
    """
    if len(seq) <= 1:
        yield seq
        yield []
    else:
        for item in powerset(seq[1:]):
            yield [seq[0]] + item
            yield item

class test_graph(TestCase):

    @dec.skipif(sys.version_info.major==3)
    def test_graph(self):
        try:
            import pydot
        except ImportError:
            raise nose.SkipTest
        mu, N, z1, z2, y = mymodel()
        for mods in [[mu], [mu, N], [mu, N, z1, z2], [mu, N, z1, z2, y]]:
            for args in powerset([('collapse_deterministics', True), ('collapse_potentials', True), ('label_edges', False), ('legend', True), ('consts', True)]):
                M = pm.Model(mods)
                pm.graph.graph(M, path=DIR, **dict(args))
                
    @dec.skipif(sys.version_info.major==3)
    def test_moral(self):
        try:
            import pydot
        except ImportError:
            raise nose.SkipTest
        mu, N, z1, z2, y = mymodel()
        for mods in [[mu], [mu, N], [mu, N, z1, z2], [mu, N, z1, z2, y]]:
            M = pm.Model(mods)
            pm.graph.moral_graph(M, path=DIR)


if __name__ == '__main__':
    C = nose.config.Config(verbosity=1)
    nose.runmodule(config=C)
